
import io
from aardwolf.protocol.x224.constants import TPDU_TYPE

# https://www.itu.int/rec/T-REC-X.224-199511-I/en (13.7)
class Data:
	def __init__(self):
		# fixed
		self.length = None #here it's always 2
		self.DT:TPDU_TYPE = TPDU_TYPE.DATA
		self.ROA:bool = False #indicates we want ACK for this packet
		self.TPDU_NR = b'\x80'
		self.data = None #remaining data which is not indicated by the length field (most of the data)
		
	def to_bytes(self):
		t = (self.DT.value << 4 | (int(self.ROA) & 0b0001)).to_bytes(1, byteorder='big', signed = False)
		t += self.TPDU_NR
		t = b'\x02' + t + self.data
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return Data.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		# be careful, this consumes the enitre buffer!
		msg = Data()
		msg.length = int.from_bytes(buff.read(1), byteorder='big', signed = False)
		t = buff.read(msg.length)
		msg.DT = TPDU_TYPE(t[0] >> 4)
		msg.ROA = bool(t[0] & 0b0001)
		msg.TPDU_NR = t[1]
		msg.data = buff.read()
		return msg

	def __repr__(self):
		t = '==== Data ====\r\n'
		t += 'length :%s\r\n' % self.length
		t += 'DT :%s\r\n' % self.DT
		t += 'ROA :%s\r\n' % self.ROA
		t += 'data :%s\r\n' % self.data
		return t


import enum
import io

class PDUTYPE(enum.Enum):
	DEMANDACTIVEPDU = 0x1 #Demand Active PDU (section 2.2.1.13.1).
	CONFIRMACTIVEPDU = 0x3 # Confirm Active PDU (section 2.2.1.13.2).
	DEACTIVATEALLPDU = 0x6 # Deactivate All PDU (section 2.2.3.1)
	DATAPDU = 0x7 # Data PDU (actual type is revealed by the pduType2 field in the Share Data Header (section 2.2.8.1.1.1.2) structure).
	SERVER_REDIR_PKT = 0xA # Enhanced Security Server Redirection PDU (section 2.2.13.3.1).

class STREAM_TYPE(enum.Enum):
	UNDEFINED = 0x00 #Undefined stream priority. This value might be used in the Server Synchronize PDU (section 2.2.1.19) due to a server-side RDP bug. It MUST NOT be used in conjunction with any other PDUs.
	LOW = 0x01 #Low-priority stream.
	MED = 0x02 #Medium-priority stream.
	HI = 0x04 #High-priority stream.

class PDUTYPE2(enum.Enum):
	UPDATE = 0x02 #Graphics Update PDU (section 2.2.9.1.1.3)
	CONTROL = 0x14 #Control PDU (section 2.2.1.15.1)
	POINTER = 0x1B #Pointer Update PDU (section 2.2.9.1.1.4)
	INPUT = 0x1C #Input Event PDU (section 2.2.8.1.1.3)
	SYNCHRONIZE = 0x1F #Synchronize PDU (section 2.2.1.14.1)
	REFRESH_RECT = 0x21 #Refresh Rect PDU (section 2.2.11.2.1)
	PLAY_SOUND = 0x22 #Play Sound PDU (section 2.2.9.1.1.5.1)
	SUPPRESS_OUTPUT = 0x23 #Suppress Output PDU (section 2.2.11.3.1)
	SHUTDOWN_REQUEST = 0x24 #Shutdown Request PDU (section 2.2.2.1.1)
	SHUTDOWN_DENIED = 0x25 #Shutdown Request Denied PDU (section 2.2.2.2.1)
	SAVE_SESSION_INFO = 0x26 #Save Session Info PDU (section 2.2.10.1.1)
	FONTLIST = 0x27 #Font List PDU (section 2.2.1.18.1)
	FONTMAP = 0x28 #Font Map PDU (section 2.2.1.22.1)
	SET_KEYBOARD_INDICATORS = 0x29 #Set Keyboard Indicators PDU (section 2.2.8.2.1.1)
	BITMAPCACHE_PERSISTENT_LIST = 0x2B #Persistent Key List PDU (section 2.2.1.17.1)
	BITMAPCACHE_ERROR_PDU = 0x2C #Bitmap Cache Error PDU ([MS-RDPEGDI] section 2.2.2.3.1)
	SET_KEYBOARD_IME_STATUS = 0x2D #Set Keyboard IME Status PDU (section 2.2.8.2.2.1)
	OFFSCRCACHE_ERROR_PDU = 0x2E #Offscreen Bitmap Cache Error PDU ([MS-RDPEGDI] section 2.2.2.3.2)
	SET_ERROR_INFO_PDU = 0x2F #Set Error Info PDU (section 2.2.5.1.1)
	DRAWNINEGRID_ERROR_PDU = 0x30 #DrawNineGrid Cache Error PDU ([MS-RDPEGDI] section 2.2.2.3.3)
	DRAWGDIPLUS_ERROR_PDU = 0x31 #GDI+ Error PDU ([MS-RDPEGDI] section 2.2.2.3.4)
	ARC_STATUS_PDU = 0x32 #Auto-Reconnect Status PDU (section 2.2.4.1.1)
	STATUS_INFO_PDU = 0x36 #Status Info PDU (section 2.2.5.2)
	MONITOR_LAYOUT_PDU = 0x37 #Monitor Layout PDU (section 2.2.12.1)

class CompType(enum.IntFlag):
	CompressionTypeMask = 0x0F #Indicates the package which was used for compression. See the table which follows for a list of compression packages.
	PACKET_COMPRESSED = 0x20 #The payload data is compressed. This flag is equivalent to MPPC bit C (for more information see [RFC2118] section 3.1).
	PACKET_AT_FRONT = 0x40 #The decompressed packet MUST be placed at the beginning of the history buffer. This flag is equivalent to MPPC bit B (for more information see [RFC2118] section 3.1).
	PACKET_FLUSHED = 0x80 #The decompressor MUST reinitialize the history buffer (by filling it with zeros) and reset the HistoryOffset to zero. After it has been reinitialized, the entire history buffer is immediately regarded as valid. This flag is equivalent to MPPC bit A (for more information see [RFC2118] section 3.1). If the PACKET_COMPRESSED (0x20) flag is also present, then the PACKET_FLUSHED flag MUST be processed first.
	PACKET_COMPR_TYPE_8K = 0x0 #RDP 4.0 bulk compression (section 3.1.8.4.1).
	PACKET_COMPR_TYPE_64K = 0x1 #RDP 5.0 bulk compression (section 3.1.8.4.2).
	PACKET_COMPR_TYPE_RDP6 = 0x2 #RDP 6.0 bulk compression ([MS-RDPEGDI] section 3.1.8.1).
	PACKET_COMPR_TYPE_RDP61 = 0x3 #RDP 6.1 bulk compression ([MS-RDPEGDI] section 3.1.8.2).

class TS_SHARECONTROLHEADER:
	def __init__(self):
		self.totalLength:int = None
		self.pduType:PDUTYPE = None
		self.pduVersion:int = 1
		self.pduSource:int = None #channel ID

	def to_bytes(self):
		t = self.totalLength.to_bytes(2, byteorder='little', signed = False)
		t += (self.pduType.value  | self.pduVersion << 4).to_bytes(2, byteorder='little', signed = False)
		t += self.pduSource.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SHARECONTROLHEADER.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SHARECONTROLHEADER()
		msg.totalLength = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		t = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.pduType = PDUTYPE(t & 0xF)
		msg.pduVersion = t >> 4
		buff.read(1)
		msg.pduSource = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_SHARECONTROLHEADER ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_SHAREDATAHEADER:
	def __init__(self):
		self.shareControlHeader:TS_SHARECONTROLHEADER = None
		self.shareID:int = None
		self.pad1:int = 0
		self.streamID:STREAM_TYPE = None
		self.uncompressedLength:int = None
		self.pduType2:PDUTYPE2 = None 
		self.compressedType:CompType = None
		self.compressedLength:int = None

	def to_bytes(self):
		t = self.shareControlHeader.to_bytes()
		t += self.shareID.to_bytes(4, byteorder='little', signed = False)
		t += self.pad1.to_bytes(1, byteorder='little', signed = False)
		t += self.streamID.value.to_bytes(1, byteorder='little', signed = False)
		t += self.uncompressedLength.to_bytes(2, byteorder='little', signed = False)
		t += self.pduType2.value.to_bytes(1, byteorder='little', signed = False)
		t += self.compressedType.to_bytes(1, byteorder='little', signed = False)
		t += self.compressedLength.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SHAREDATAHEADER.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SHAREDATAHEADER()
		msg.shareControlHeader = TS_SHARECONTROLHEADER.from_buffer(buff)
		msg.shareID = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.pad1 = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.streamID = STREAM_TYPE(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		msg.uncompressedLength = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.pduType2 = PDUTYPE2(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		msg.compressedType = CompType(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		msg.compressedLength = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_SHAREDATAHEADER ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum
from aardwolf.protocol.T128.share import TS_SHAREDATAHEADER

class CTRLACTION(enum.Enum):
	REQUEST_CONTROL = 0x0001 #Request control
	GRANTED_CONTROL = 0x0002 #Granted control
	DETACH = 0x0003 #Detach
	COOPERATE = 0x0004 #Cooperate

class TS_CONTROL_PDU:
	def __init__(self):
		self.shareDataHeader:TS_SHAREDATAHEADER = None
		self.action:CTRLACTION = None
		self.grantId:int = 0
		self.controlId:int = 0

	def to_bytes(self):
		#t  = self.shareDataHeader.to_bytes()
		t = self.action.value.to_bytes(2, byteorder='little', signed = False)
		t += self.grantId.to_bytes(2, byteorder='little', signed = False)
		t += self.controlId.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_CONTROL_PDU.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_CONTROL_PDU()
		msg.shareDataHeader = TS_SHAREDATAHEADER.from_buffer(buff)
		msg.action = CTRLACTION(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.grantId = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.controlId = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_CONTROL_PDU ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum
from typing import Any, List

class CAPSTYPE(enum.Enum):
	GENERAL = 0x0001 #General Capability Set

class CB_CAPS_VERSION(enum.Enum):
	VERSION_1 = 0x00000001 #Version 1
	VERSION_2 = 0x00000002 #Version 2

class CB_GENERAL_FALGS(enum.IntFlag):
	USE_LONG_FORMAT_NAMES = 0x00000002 #The Long Format Name variant of the Format List PDU is supported for exchanging updated format names. If this flag is not set, the Short Format Name variant MUST be used. If this flag is set by both protocol endpoints, then the Long Format Name variant MUST be used.
	STREAM_FILECLIP_ENABLED = 0x00000004 #File copy and paste using stream-based operations are supported using the File Contents Request PDU and File Contents Response PDU.
	FILECLIP_NO_FILE_PATHS = 0x00000008 #Indicates that any description of files to copy and paste MUST NOT include the source path of the files.
	CAN_LOCK_CLIPDATA = 0x00000010 #Locking and unlocking of File Stream data on the clipboard is supported using the Lock Clipboard Data PDU and Unlock Clipboard Data PDU.
	HUGE_FILE_SUPPORT_ENABLED = 0x00000020 #Indicates support for transferring files that are larger than 4,294,967,295 bytes in size. If this flag is not set, then only files of size less than or equal to 4,294,967,295 bytes can be exchanged using the File Contents Request PDU and File Contents Response PDU.

class CLIPRDR_GENERAL_CAPABILITY:
	def __init__(self):
		self.capabilitySetType:CAPSTYPE = CAPSTYPE.GENERAL
		self.lengthCapability:int = 12
		self.version:CB_CAPS_VERSION = CB_CAPS_VERSION.VERSION_2
		self.generalFlags:CB_GENERAL_FALGS = 0

	def to_bytes(self):
		t = self.capabilitySetType.value.to_bytes(2, byteorder='little', signed = False)
		t += self.lengthCapability.to_bytes(2, byteorder='little', signed = False)
		t += self.version.value.to_bytes(4, byteorder='little', signed = False)
		t += self.generalFlags.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_GENERAL_CAPABILITY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_GENERAL_CAPABILITY()
		msg.capabilitySetType = CAPSTYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.lengthCapability = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.version = CB_CAPS_VERSION(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.generalFlags = CB_GENERAL_FALGS(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_GENERAL_CAPABILITY ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


class CLIPRDR_CAPS:
	def __init__(self):
		self.cCapabilitiesSets:int = None
		self.pad1:bytes = b'\x00\x00'
		self.capabilitySets:List[Any] = []


	def to_bytes(self):
		t = len(self.capabilitySets).to_bytes(2, byteorder='little', signed = False)
		t += self.pad1
		for cap in self.capabilitySets:
			t += cap.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_CAPS.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_CAPS()
		msg.cCapabilitiesSets = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.pad1 = buff.read(2)
		for _ in range(msg.cCapabilitiesSets):
			obj = capstype2cap[CAPSTYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))]
			buff.seek(-2,1)
			msg.capabilitySets.append(obj.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_CAPS ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

capstype2cap = {
	CAPSTYPE.GENERAL : CLIPRDR_GENERAL_CAPABILITY
}


import asyncio

from aardwolf.extensions.RDPEDYC.protocol import DYNVC_CMD
from aardwolf.extensions.RDPEDYC.protocol.data import DYNVC_DATA_FIRST, DYNVC_DATA

class VirtualChannelBase:
	def __init__(self, name:str, buffered = True):
		self.channel_name:str = name
		self.channel_id:int = None
		self.channel_closed_evt:asyncio.Event = asyncio.Event()
		self.__channel_data_buffered = buffered
		self.__virtual_channel_manager = None
		self.__fragment_buffer:bytes = b''
		self.__current_data_length = -1
		self.__send_lock = asyncio.Lock()

	async def channel_init_internal(self, channelid:int, manager):
		self.channel_id = channelid
		self.__virtual_channel_manager = manager
		return await self.channel_init()

	async def channel_rawdata_in(self, msg:DYNVC_DATA_FIRST or DYNVC_DATA):
		"""
		This method is called by the Channel Manager whenever (uncompressed) data is coming in.
		The data might be fragmented, so this function serves to reconstruct the entire data that is
		send by the server. 

		"""
		if self.__channel_data_buffered is False:
			await self.channel_data_in(msg.Data)
			return
		
		self.__fragment_buffer += msg.Data
		if msg.cmd == DYNVC_CMD.DATA_FIRST:
			self.__current_data_length = msg.Length

		if len(self.__fragment_buffer) == self.__current_data_length or self.__current_data_length == -1:
			await self.channel_data_in(self.__fragment_buffer)
			self.__fragment_buffer = b''
			self.__current_data_length = -1
		#print('__current_data_length: %s' % self.__current_data_length)
		#print('len(self.__fragment_buffer): %s' % len(self.__fragment_buffer))

	async def channel_data_out(self, data:bytes):
		async with self.__send_lock:
			for datamsg in DYNVC_DATA_FIRST.chunk_data(data, self.channel_id):
				#print('Sending reply data: %s' % datamsg)
				await self.__virtual_channel_manager.fragment_and_send(datamsg)
	
	async def channel_closed(self):
		"""
		The remote end closed the channel
		"""
		self.channel_closed_evt.set()

	async def close_channel(self):
		"""
		Triggers the channel to be closed
		"""
		await self.__virtual_channel_manager.close_channel(self.channel_id)

	#### OVERRIDE THESE
	async def channel_init(self):
		"""
		This function is called when the channel is about to start.
		YOU MUST return a Tuple[Bool, Exception] after finishing init.
		If you return exception in the tuple, the channel will be discarded.
		"""
		return True, None

	async def channel_data_in(self, data:bytes):
		"""
		This function will be called when data arrives on this channel.
		If you used unbuffered channel, the data here will arrive in 1600 byte chunks
		"""
		pass

	

	



import io
import enum
from aardwolf.protocol.T128.share import TS_SHAREDATAHEADER

class TS_FONT_LIST_PDU:
	def __init__(self):
		self.shareDataHeader:TS_SHAREDATAHEADER = None
		self.numberFonts:int = 0
		self.totalNumFonts:int = 0
		self.listFlags:int = 0x0003
		self.entrySize:int = 50

	def to_bytes(self):
		#t  = self.shareDataHeader.to_bytes()
		t = self.numberFonts.to_bytes(2, byteorder='little', signed = False)
		t += self.totalNumFonts.to_bytes(2, byteorder='little', signed = False)
		t += self.listFlags.to_bytes(2, byteorder='little', signed = False)
		t += self.entrySize.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FONT_LIST_PDU.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FONT_LIST_PDU()
		msg.shareDataHeader = TS_SHAREDATAHEADER.from_buffer(buff)
		msg.numberFonts = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.totalNumFonts = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.listFlags = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.entrySize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_FONT_LIST_PDU ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import io
import enum
from typing import List

from aardwolf.extensions.RDPECLIP.protocol.formatlist import CLIPBRD_FORMAT

class MAPPING_MODE(enum.Enum):
	TEXT = 0x00000001 #Each logical unit is mapped to one device pixel. Positive x is to the right; positive y is down.
	LOMETRIC = 0x00000002 #Each logical unit is mapped to 0.1 millimeter. Positive x is to the right; positive y is up.
	HIMETRIC = 0x00000003 #Each logical unit is mapped to 0.01 millimeter. Positive x is to the right; positive y is up.
	LOENGLISH = 0x00000004 #Each logical unit is mapped to 0.01 inch. Positive x is to the right; positive y is up.
	HIENGLISH = 0x00000005 #Each logical unit is mapped to 0.001 inch. Positive x is to the right; positive y is up.
	TWIPS = 0x00000006 #Each logical unit is mapped to 1/20 of a printer's point (1/1440 of an inch), also called a twip. Positive x is to the right; positive y is up.
	ISOTROPIC = 0x00000007 #Logical units are mapped to arbitrary units with equally scaled axes; one unit along the x-axis is equal to one unit along the y-axis.
	ANISOTROPIC = 0x00000008


class CLIPRDR_MFPICT:
	def __init__(self):
		self.mappingMode:MAPPING_MODE = None
		self.xExt:int = None
		self.yExt:int = None
		self.metaFileData:bytes = None

	def to_bytes(self):
		t = self.mappingMode.value.to_bytes(4, byteorder='little', signed=False)
		t += self.xExt.to_bytes(4, byteorder='little', signed=False)
		t += self.yExt.to_bytes(4, byteorder='little', signed=False)
		t += self.metaFileData
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_MFPICT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_MFPICT()
		msg.mappingMode = MAPPING_MODE(int.from_bytes(buff.read(4), byteorder='little', signed=False))
		msg.xExt = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		msg.yExt = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		msg.metaFileData = buff.read()
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_MFPICT ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class PALETTEENTRY:
	def __init__(self):
		self.red:int = None
		self.green:int = None
		self.blue:int = None
		self.extra:bytes = b'\x00'

	def to_bytes(self):
		t = self.red.to_bytes(1, byteorder='little', signed=False)
		t += self.green.to_bytes(1, byteorder='little', signed=False)
		t += self.blue.to_bytes(1, byteorder='little', signed=False)
		t += self.extra
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return PALETTEENTRY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = PALETTEENTRY()
		msg.red = buff.read(1)[0]
		msg.green = buff.read(1)[0]
		msg.blue = buff.read(1)[0]
		msg.extra = buff.read(1)
		return msg

	def __repr__(self):
		t = '==== PALETTEENTRY ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


class CLIPRDR_PALETTE:
	def __init__(self):
		self.paletteEntriesData:List[PALETTEENTRY] = []

	def to_bytes(self):
		t = b''
		for pe in self.paletteEntriesData:
			t += pe.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_PALETTE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_PALETTE()
		while buff.read(1) != b'':
			buff.seek(-1,1)
			PALETTEENTRY.from_buffer(buff)
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_PALETTE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class FD_FLAGS(enum.IntFlag):
	ATTRIBUTES = 0x00000004 #The fileAttributes field contains valid data.
	FILESIZE = 0x00000040 #The fileSizeHigh and fileSizeLow fields contain valid data.
	WRITESTIME = 0x00000020 #The lastWriteTime field contains valid data.
	SHOWPROGRESSUI = 0x00004000 #A progress indicator SHOULD be shown when copying the file.

class FILE_ATTRIBUTE(enum.IntFlag):
	READONLY = 0x00000001 #A file that is read-only. Applications can read the file, but cannot write to it or delete it.
	HIDDEN = 0x00000002 #The file or directory is hidden. It is not included in an ordinary directory listing.
	SYSTEM = 0x00000004 #A file or directory that the operating system uses a part of, or uses exclusively.
	DIRECTORY = 0x00000010 #Identifies a directory.
	ARCHIVE = 0x00000020 #A file or directory that is an archive file or directory. Applications typically use this attribute to mark files for backup or removal.
	NORMAL = 0x00000080 #A file that does not have other attributes set. This attribute is valid only when used alone.

class CLIPRDR_FILEDESCRIPTOR:
	def __init__(self):
		self.flags:FD_FLAGS = None
		self.reserved1:bytes = b'\x00'*32
		self.fileAttributes:FILE_ATTRIBUTE = None
		self.reserved2:bytes = None
		self.lastWriteTime:int = None
		self.fileSizeHigh:int = None
		self.fileSizeLow:int = None
		self.fileName:str = None #520 bytes, 260 chars null term

		self.fileSize:int =None

	def to_bytes(self):
		if self.fileSize is not None:
			self.fileSizeHigh = self.fileSize >> 32
			self.fileSizeLow = self.fileSize & 0xFFFFFFFF

		t = self.flags.to_bytes(4, byteorder='little', signed=False)
		t += self.reserved1
		t += self.fileAttributes.to_bytes(4, byteorder='little', signed=False)
		t += self.reserved2
		t += self.lastWriteTime.to_bytes(8, byteorder='little', signed=False)
		t += self.fileSizeHigh.to_bytes(4, byteorder='little', signed=False)
		t += self.fileSizeLow.to_bytes(4, byteorder='little', signed=False)
		t += self.fileName.encode('utf-16-le').ljust(b'\x00',520)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_FILEDESCRIPTOR.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_FILEDESCRIPTOR()
		msg.flags = FD_FLAGS(int.from_bytes(buff.read(4), byteorder='little', signed=False))
		msg.reserved1 = buff.read(32)
		msg.fileAttributes = FILE_ATTRIBUTE(int.from_bytes(buff.read(4), byteorder='little', signed=False))
		msg.reserved2 = buff.read(16)
		msg.lastWriteTime = FILE_ATTRIBUTE(int.from_bytes(buff.read(8), byteorder='little', signed=False))
		msg.fileSizeHigh = FILE_ATTRIBUTE(int.from_bytes(buff.read(4), byteorder='little', signed=False))
		msg.fileSizeLow = FILE_ATTRIBUTE(int.from_bytes(buff.read(4), byteorder='little', signed=False))
		msg.fileName = buff.read(520).decode('utf-16-le').replace('\x00','')
		msg.fileSize = msg.fileSizeHigh << 32 | msg.fileSizeLow
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_FILEDESCRIPTOR ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


class CLIPRDR_FILELIST:
	def __init__(self):
		self.cItems:MAPPING_MODE = None
		self.fileDescriptorArray:List[CLIPRDR_FILEDESCRIPTOR] = []

	def to_bytes(self):
		t = len(self.fileDescriptorArray).to_bytes(4, byteorder='little', signed=False)
		for fd in self.fileDescriptorArray:
			t += fd.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return CLIPRDR_FILELIST.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = CLIPRDR_FILELIST()
		msg.cItems = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		for _ in range(msg.cItems):
			msg.fileDescriptorArray.append(CLIPRDR_FILEDESCRIPTOR.from_buffer(buff))
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_FILELIST ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class CLIPRDR_FORMAT_DATA_RESPONSE:
	def __init__(self):
		self.dataobj = None

	def to_bytes(self, otype:CLIPBRD_FORMAT):
		if otype == CLIPBRD_FORMAT.CF_UNICODETEXT:
			t = self.dataobj.encode('utf-16-le') + b'\x00'*3
		elif otype == CLIPBRD_FORMAT.CF_TEXT:
			t = self.dataobj.encode('ascii') + b'\x00'
		else:
			t = self.dataobj.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes, otype:CLIPBRD_FORMAT):
		return CLIPRDR_FORMAT_DATA_RESPONSE.from_buffer(io.BytesIO(bbuff), otype = otype)

	@staticmethod
	def from_buffer(buff: io.BytesIO, otype:CLIPBRD_FORMAT):
		msg = CLIPRDR_FORMAT_DATA_RESPONSE()
		if otype == CLIPBRD_FORMAT.CF_HDROP: #TODO: not sure
			msg.dataobj = CLIPRDR_FILELIST.from_buffer(buff)
		elif otype == CLIPBRD_FORMAT.CF_PALETTE:
			msg.dataobj =  CLIPRDR_PALETTE.from_buffer(buff)
		elif otype == CLIPBRD_FORMAT.CF_METAFILEPICT:
			msg.dataobj =  CLIPRDR_MFPICT.from_buffer(buff)
		elif otype == CLIPBRD_FORMAT.CF_UNICODETEXT:
			msg.dataobj = buff.read().decode('utf-16-le').replace('\x00','')
		elif otype == CLIPBRD_FORMAT.CF_TEXT:
			msg.dataobj = buff.read().decode().replace('\x00','')
		return msg

	def __repr__(self):
		t = '==== CLIPRDR_FORMAT_DATA_RESPONSE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

# again, this is so messed up. The same commandid can be request or response as well...
class DYNVC_CMD(enum.Enum):
    CREATE_RSP = 0x01 #The message contained in the optionalFields field is a Create Request PDU (section 2.2.2.1) or a Create Response PDU (section 2.2.2.2).
    DATA_FIRST = 0x02 #The message contained in the optionalFields field is a Data First PDU (section 2.2.3.1).
    DATA = 0x03 #The message contained in the optionalFields field is a Data PDU (section 2.2.3.2).
    CLOSE = 0x04 #The message contained in the optionalFields field is a Close Request PDU (section 2.2.4) or a Close Response PDU (section 2.2.4).
    CAPS_RSP = 0x05 #The message contained in the optionalFields field is a Capability Request PDU (section 2.2.1.1) or a Capabilities Response PDU (section 2.2.1.2).
    DATA_FIRST_COMPRESSED = 0x06 #The message contained in the optionalFields field is a Data First Compressed PDU (section 2.2.3.3).
    DATA_COMPRESSED = 0x07 #The message contained in the optionalFields field is a Data Compressed PDU (section 2.2.3.4).
    SOFT_SYNC_REQUEST = 0x08 #The message contained in the optionalFields field is a Soft-Sync Request PDU (section 2.2.5.1).
    SOFT_SYNC_RESPONSE = 0x09 #The message contained in the optionalFields field is a Soft-Sync Response PDU (section 2.2.5.2).


def dynvc_header_from_bytes(data:bytes, cbid_mod = False, sp_mod = False):
    hdr = data[0]
    cmd = DYNVC_CMD(hdr >> 4)
    sp = (hdr >> 2) & 0b11
    cbid = hdr & 0b11
    if cbid_mod is True:
        if cbid == 0: cbid += 1
        else: cbid = cbid * 2
    if sp_mod is True:
        if sp == 0: sp += 1
        else: sp = sp * 2
    return cbid, sp, cmd

def dynvc_header_from_buff(buffer:io.BytesIO, cbid_mod = False, sp_mod = False):
    return dynvc_header_from_bytes(buffer.read(1), cbid_mod=cbid_mod, sp_mod=sp_mod)

def dynvc_header_to_bytes(cbid:int, sp:int, cmd:DYNVC_CMD, cbid_mod = False, sp_mod = False):
    if cbid_mod is True:
        if cbid == 1: cbid = 0
        elif cbid == 2: cbid = 1
        elif cbid == 4: cbid = 2
    if sp_mod is True:
        if sp == 1: sp = 0
        elif sp == 2: sp = 1
        elif sp == 4: sp = 2
    return bytes([(cmd.value << 4) ^ (sp << 2) ^ cbid]) 

from aardwolf.extensions.RDPEDYC.protocol.softsync import DYNVC_SOFT_SYNC_REQUEST, DYNVC_SOFT_SYNC_RESPONSE
from aardwolf.extensions.RDPEDYC.protocol.close import DYNVC_CLOSE
from aardwolf.extensions.RDPEDYC.protocol.data import DYNVC_DATA_FIRST, DYNVC_DATA
from aardwolf.extensions.RDPEDYC.protocol.create import DYNVC_CREATE_REQ, DYNVC_CREATE_RSP
from aardwolf.extensions.RDPEDYC.protocol.caps import DYNVC_CAPS_REQ

class DYNVC_MESSAGE:
    def __init__(self):
        self.cmd:DYNVC_CMD2MSG = None

    @staticmethod
    def from_bytes(data:bytes):
        return DYNVC_MESSAGE.from_buffer(io.BytesIO(data))
    
    @staticmethod
    def from_buffer(buff:io.BytesIO):
        _, _, cmd = dynvc_header_from_bytes(buff.read(1))
        buff.seek(-1, io.SEEK_CUR)
        if DYNVC_CMD2MSG[cmd] is None:
            raise NotImplementedError()
        return DYNVC_CMD2MSG[cmd].from_buffer(buff)
        

DYNVC_CMD2MSG = {
    DYNVC_CMD.CREATE_RSP : DYNVC_CREATE_REQ,
    DYNVC_CMD.DATA_FIRST : DYNVC_DATA_FIRST,
    DYNVC_CMD.DATA : DYNVC_DATA,
    DYNVC_CMD.CLOSE : DYNVC_CLOSE,
    DYNVC_CMD.CAPS_RSP : DYNVC_CAPS_REQ,
    DYNVC_CMD.DATA_FIRST_COMPRESSED : None, #not supported!
    DYNVC_CMD.DATA_COMPRESSED : None, # not supported!
    DYNVC_CMD.SOFT_SYNC_REQUEST : DYNVC_SOFT_SYNC_REQUEST,
    DYNVC_CMD.SOFT_SYNC_RESPONSE : DYNVC_SOFT_SYNC_RESPONSE,
}


import enum
import io
from aardwolf.protocol.T128.share import TS_SHAREDATAHEADER

class ERRINFO(enum.Enum):
	UNK = 'blabla' # error not found
	NONE = 0x00000000 # No error has occurred. This code SHOULD be ignored.
	RPC_INITIATED_DISCONNECT = 0x00000001 # The disconnection was initiated by an administrative tool on the server in another session.
	RPC_INITIATED_LOGOFF = 0x00000002 # The disconnection was due to a forced logoff initiated by an administrative tool on the server in another session.
	IDLE_TIMEOUT = 0x00000003 # The idle session limit timer on the server has elapsed.
	LOGON_TIMEOUT = 0x00000004 # The active session limit timer on the server has elapsed.
	DISCONNECTED_BY_OTHERCONNECTION = 0x00000005 # Another user connected to the server, forcing the disconnection of the current connection.
	OUT_OF_MEMORY = 0x00000006 # The server ran out of available memory resources. 
	SERVER_DENIED_CONNECTION = 0x00000007 # The server denied the connection.
	SERVER_INSUFFICIENT_PRIVILEGES = 0x00000009 # The user cannot connect to the server due to insufficient access privileges.
	SERVER_FRESH_CREDENTIALS_REQUIRED = 0x0000000A # The server does not accept saved user credentials and requires that the user enter their credentials for each connection.
	RPC_INITIATED_DISCONNECT_BYUSER = 0x0000000B # The disconnection was initiated by an administrative tool on the server running in the user’s session.
	LOGOFF_BY_USER = 0x0000000C # The disconnection was initiated by the user logging off his or her session on the server.
	CLOSE_STACK_ON_DRIVER_NOT_READY = 0x0000000F # The display driver in the remote session did not report any status within the time allotted for startup.
	SERVER_DWM_CRASH = 0x00000010 # The DWM process running in the remote session terminated unexpectedly.
	CLOSE_STACK_ON_DRIVER_FAILURE = 0x00000011 # The display driver in the remote session was unable to complete all the tasks required for startup.
	CLOSE_STACK_ON_DRIVER_IFACE_FAILURE = 0x00000012 # The display driver in the remote session started up successfully, but due to internal failures was not usable by the remoting stack.
	SERVER_WINLOGON_CRASH = 0x00000017 # The Winlogon process running in the remote session terminated unexpectedly.
	SERVER_CSRSS_CRASH = 0x00000018 # The CSRSS process running in the remote session terminated unexpectedly.
	SERVER_SHUTDOWN = 0x00000019 # The remote server is busy shutting down.
	SERVER_REBOOT = 0x0000001A # The remote server is busy rebooting.
	LICENSE_INTERNAL = 0x00000100 # An internal error has occurred in the Terminal Services licensing component.
	LICENSE_NO_LICENSE_SERVER = 0x00000101 # A Remote Desktop License Server ([MS-RDPELE] section 1.1) could not be found to provide a license.
	LICENSE_NO_LICENSE = 0x00000102 # There are no Client Access Licenses ([MS-RDPELE] section 1.1) available for the target remote computer.
	LICENSE_BAD_CLIENT_MSG = 0x00000103 # The remote computer received an invalid licensing message from the client.
	LICENSE_HWID_DOESNT_MATCH_LICENSE = 0x00000104 # The Client Access License ([MS-RDPELE] section 1.1) stored by the client has been modified.
	LICENSE_BAD_CLIENT_LICENSE = 0x00000105 # The Client Access License ([MS-RDPELE] section 1.1) stored by the client is in an invalid format
	LICENSE_CANT_FINISH_PROTOCOL = 0x00000106 # Network problems have caused the licensing protocol ([MS-RDPELE] section 1.3.3) to be terminated.
	LICENSE_CLIENT_ENDED_PROTOCOL = 0x00000107 #The client prematurely ended the licensing protocol ([MS-RDPELE] section 1.3.3).
	LICENSE_BAD_CLIENT_ENCRYPTION = 0x00000108 #A licensing message ([MS-RDPELE] sections 2.2 and 5.1) was incorrectly encrypted.
	LICENSE_CANT_UPGRADE_LICENSE = 0x00000109 # The Client Access License ([MS-RDPELE] section 1.1) stored by the client could not be upgraded or renewed.
	LICENSE_NO_REMOTE_CONNECTIONS = 0x0000010A # The remote computer is not licensed to accept remote connections.
	CB_DESTINATION_NOT_FOUND = 0x00000400 #The target endpoint could not be found.
	CB_LOADING_DESTINATION = 0x00000402 #The target endpoint to which the client is being redirected is disconnecting from the Connection Broker.
	CB_REDIRECTING_TO_DESTINATION = 0x00000404 #An error occurred while the connection was being redirected to the target endpoint.
	CB_SESSION_ONLINE_VM_WAKE = 0x00000405 #An error occurred while the target endpoint (a virtual machine) was being awakened.
	CB_SESSION_ONLINE_VM_BOOT = 0x00000406 #An error occurred while the target endpoint (a virtual machine) was being started.
	CB_SESSION_ONLINE_VM_NO_DNS = 0x00000407 #The IP address of the target endpoint (a virtual machine) cannot be determined.
	CB_DESTINATION_POOL_NOT_FREE = 0x00000408 #There are no available endpoints in the pool managed by the Connection Broker.
	CB_CONNECTION_CANCELLED = 0x00000409 #Processing of the connection has been canceled.
	CB_CONNECTION_ERROR_INVALID_SETTINGS = 0x00000410 #The settings contained in the routingToken field of the X.224 Connection Request PDU (section 2.2.1.1) cannot be validated.
	CB_SESSION_ONLINE_VM_BOOT_TIMEOUT = 0x00000411 #A time-out occurred while the target endpoint (a virtual machine) was being started.
	CB_SESSION_ONLINE_VM_SESSMON_FAILED = 0x00000412 #A session monitoring error occurred while the target endpoint (a virtual machine) was being started.
	UNKNOWNPDUTYPE2 = 0x000010C9 #Unknown pduType2 field in a received Share Data Header (section 2.2.8.1.1.1.2). 
	UNKNOWNPDUTYPE = 0x000010CA #Unknown pduType field in a received Share Control Header (section 2.2.8.1.1.1.1).
	DATAPDUSEQUENCE = 0x000010CB #An out-of-sequence Slow-Path Data PDU (section 2.2.8.1.1.1.1) has been received.
	CONTROLPDUSEQUENCE = 0x000010CD #An out-of-sequence Demand Active PDU (section 2.2.1.13.1), Confirm Active PDU (section 2.2.1.13.2), Deactivate All PDU (section 2.2.3.1) or Enhanced Security Server Redirection PDU (section 2.2.13.3.1) has been received.
	INVALIDCONTROLPDUACTION = 0x000010CE #A Control PDU (sections 2.2.1.15 and 2.2.1.16) has been received with an invalid action field.
	INVALIDINPUTPDUTYPE = 0x000010CF #One of two possible errors #    A Slow-Path Input Event (section 2.2.8.1.1.3.1.1) has been received with an invalid messageType field. #A Fast-Path Input Event (section 2.2.8.1.2.2) has been received with an invalid eventCode field.
	INVALIDINPUTPDUMOUSE = 0x000010D0 #One of two possible errors: #    A Slow-Path Mouse Event (section 2.2.8.1.1.3.1.1.3) or Extended Mouse Event (section 2.2.8.1.1.3.1.1.4) has been received with an invalid pointerFlags field #A Fast-Path Mouse Event (section 2.2.8.1.2.2.3) or Fast-Path Extended Mouse Event (section 2.2.8.1.2.2.4) has been received with an invalid pointerFlags field.
	INVALIDREFRESHRECTPDU = 0x000010D1 #An invalid Refresh Rect PDU (section 2.2.11.2) has been received.
	CREATEUSERDATAFAILED = 0x000010D2 #The server failed to construct the GCC Conference Create Response user data (section 2.2.1.4).
	CONNECTFAILED = 0x000010D3 #Processing during the Channel Connection phase of the RDP Connection Sequence (see section 1.3.1.1 for an overview of the RDP Connection Sequence phases) has failed.
	CONFIRMACTIVEWRONGSHAREID = 0x000010D4 #A Confirm Active PDU (section 2.2.1.13.2) was received from the client with an invalid shareID field.
	CONFIRMACTIVEWRONGORIGINATOR = 0x000010D5 #A Confirm Active PDU (section 2.2.1.13.2) was received from the client with an invalid originatorID field.
	PERSISTENTKEYPDUBADLENGTH = 0x000010DA #There is not enough data to process a Persistent Key List PDU (section 2.2.1.17).
	PERSISTENTKEYPDUILLEGALFIRST = 0x000010DB #A Persistent Key List PDU (section 2.2.1.17) marked as PERSIST_PDU_FIRST (0x01) was received after the reception of a prior Persistent Key List PDU also marked as PERSIST_PDU_FIRST.
	PERSISTENTKEYPDUTOOMANYTOTALKEYS = 0x000010DC #A Persistent Key List PDU (section 2.2.1.17) was received which specified a total number of bitmap cache entries larger than 262144.
	PERSISTENTKEYPDUTOOMANYCACHEKEYS = 0x000010DD #A Persistent Key List PDU (section 2.2.1.17) was received which specified an invalid total number of keys for a bitmap cache (the number of entries that can be stored within each bitmap cache is specified in the Revision 1 or 2 Bitmap Cache Capability Set (section 2.2.7.1.4) that is sent from client to server).
	INPUTPDUBADLENGTH = 0x000010DE #There is not enough data to process Input Event PDU Data (section 2.2.8.1.1.3.1) or a Fast-Path Input Event PDU (section 2.2.8.1.2).
	BITMAPCACHEERRORPDUBADLENGTH = 0x000010DF #There is not enough data to process the shareDataHeader, NumInfoBlocks, Pad1, and Pad2 fields of the Bitmap Cache Error PDU Data ([MS-RDPEGDI] section 2.2.2.3.1.1).
	SECURITYDATATOOSHORT = 0x000010E0 #One of two possible errors: #    The dataSignature field of the Fast-Path Input Event PDU (section 2.2.8.1.2) does not contain enough data #The fipsInformation and dataSignature fields of the Fast-Path Input Event PDU (section 2.2.8.1.2) do not contain enough data.
	VCHANNELDATATOOSHORT = 0x000010E1 #One of two possible errors #    There is not enough data in the Client Network Data (section 2.2.1.3.4) to read the virtual channel configuration data. #There is not enough data to read a complete Channel PDU Header (section 2.2.6.1.1).
	SHAREDATATOOSHORT = 0x000010E2 #One of four possible errors: #There is not enough data to process Control PDU Data (section 2.2.1.15.1). #There is not enough data to read a complete Share Control Header (section 2.2.8.1.1.1.1). #There is not enough data to read a complete Share Data Header (section 2.2.8.1.1.1.2) of a Slow-Path Data PDU (section 2.2.8.1.1.1.1). #There is not enough data to process Font List PDU Data (section 2.2.1.18.1).
	BADSUPRESSOUTPUTPDU = 0x000010E3 #One of two possible errors: #    There is not enough data to process Suppress Output PDU Data (section 2.2.11.3.1).#    The allowDisplayUpdates field of the Suppress Output PDU Data (section 2.2.11.3.1) is invalid.
	CONFIRMACTIVEPDUTOOSHORT = 0x000010E5 #One of two possible errors: #    There is not enough data to read the shareControlHeader, shareID, originatorID, lengthSourceDescriptor, and lengthCombinedCapabilities fields of the Confirm Active PDU Data (section 2.2.1.13.2.1).# There is not enough data to read the sourceDescriptor, numberCapabilities, pad2Octets, and capabilitySets fields of the Confirm Active PDU Data (section 2.2.1.13.2.1).
	CAPABILITYSETTOOSMALL = 0x000010E7 #There is not enough data to read the capabilitySetType and the lengthCapability fields in a received Capability Set (section 2.2.1.13.1.1.1).
	CAPABILITYSETTOOLARGE = 0x000010E8 #A Capability Set (section 2.2.1.13.1.1.1) has been received with a lengthCapability field that contains a value greater than the total length of the data received.
	NOCURSORCACHE = 0x000010E9 #One of two possible errors: #    Both the colorPointerCacheSize and pointerCacheSize fields in the Pointer Capability Set (section 2.2.7.1.5) are set to zero. #The pointerCacheSize field in the Pointer Capability Set (section 2.2.7.1.5) is not present, and the colorPointerCacheSize field is set to zero.
	BADCAPABILITIES = 0x000010EA #The capabilities received from the client in the Confirm Active PDU (section 2.2.1.13.2) were not accepted by the server.
	VIRTUALCHANNELDECOMPRESSIONERR = 0x000010EC #An error occurred while using the bulk compressor (section 3.1.8 and [MS-RDPEGDI] section 3.1.8) to decompress a Virtual Channel PDU (section 2.2.6.1)
	INVALIDVCCOMPRESSIONTYPE = 0x000010ED #An invalid bulk compression package was specified in the flags field of the Channel PDU Header (section 2.2.6.1.1).
	INVALIDCHANNELID = 0x000010EF #An invalid MCS channel ID was specified in the mcsPdu field of the Virtual Channel PDU (section 2.2.6.1).
	VCHANNELSTOOMANY = 0x000010F0 #The client requested more than the maximum allowed 31 static virtual channels in the Client Network Data (section 2.2.1.3.4).
	REMOTEAPPSNOTENABLED = 0x000010F3 #The INFO_RAIL flag (0x00008000) MUST be set in the flags field of the Info Packet (section 2.2.1.11.1.1) as the session on the remote server can only host remote applications.
	CACHECAPNOTSET = 0x000010F4 #The client sent a Persistent Key List PDU (section 2.2.1.17) without including the prerequisite Revision 2 Bitmap Cache Capability Set (section 2.2.7.1.4.2) in the Confirm Active PDU (section 2.2.1.13.2).
	BITMAPCACHEERRORPDUBADLENGTH2 = 0x000010F5 #The NumInfoBlocks field in the Bitmap Cache Error PDU Data is inconsistent with the amount of data in the Info field ([MS-RDPEGDI] section 2.2.2.3.1.1).
	OFFSCRCACHEERRORPDUBADLENGTH = 0x000010F6 #There is not enough data to process an Offscreen Bitmap Cache Error PDU ([MS-RDPEGDI] section 2.2.2.3.2).
	DNGCACHEERRORPDUBADLENGTH = 0x000010F7 #There is not enough data to process a DrawNineGrid Cache Error PDU ([MS-RDPEGDI] section 2.2.2.3.3).
	GDIPLUSPDUBADLENGTH = 0x000010F8 #There is not enough data to process a GDI+ Error PDU ([MS-RDPEGDI] section 2.2.2.3.4).
	SECURITYDATATOOSHORT2 = 0x00001111 #There is not enough data to read a Basic Security Header (section 2.2.8.1.1.2.1).
	SECURITYDATATOOSHORT3 = 0x00001112 #There is not enough data to read a Non-FIPS Security Header (section 2.2.8.1.1.2.2) or FIPS Security Header (section 2.2.8.1.1.2.3).
	SECURITYDATATOOSHORT4 = 0x00001113 #There is not enough data to read the basicSecurityHeader and length fields of the Security Exchange PDU Data (section 2.2.1.10.1).
	SECURITYDATATOOSHORT5 = 0x00001114 #There is not enough data to read the CodePage, flags, cbDomain, cbUserName, cbPassword, cbAlternateShell, cbWorkingDir, Domain, UserName, Password, AlternateShell, and WorkingDir fields in the Info Packet (section 2.2.1.11.1.1).
	SECURITYDATATOOSHORT6 = 0x00001115 #There is not enough data to read the CodePage, flags, cbDomain, cbUserName, cbPassword, cbAlternateShell, and cbWorkingDir fields in the Info Packet (section 2.2.1.11.1.1).
	SECURITYDATATOOSHORT7 = 0x00001116 #There is not enough data to read the clientAddressFamily and cbClientAddress fields in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT8 = 0x00001117 #There is not enough data to read the clientAddress field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT9 = 0x00001118 #There is not enough data to read the cbClientDir field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT10 = 0x00001119 #There is not enough data to read the clientDir field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT11 = 0x0000111A #There is not enough data to read the clientTimeZone field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT12 = 0x0000111B #There is not enough data to read the clientSessionId field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT13 = 0x0000111C #There is not enough data to read the performanceFlags field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT14 = 0x0000111D #There is not enough data to read the cbAutoReconnectCookie field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT15 = 0x0000111E #There is not enough data to read the autoReconnectCookie field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT16 = 0x0000111F #The cbAutoReconnectCookie field in the Extended Info Packet (section 2.2.1.11.1.1.1) contains a value which is larger than the maximum allowed length of 128 bytes.
	SECURITYDATATOOSHORT17 = 0x00001120 #There is not enough data to read the clientAddressFamily and cbClientAddress fields in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT18 = 0x00001121 #There is not enough data to read the clientAddress field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT19 = 0x00001122 #There is not enough data to read the cbClientDir field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT20 = 0x00001123 #There is not enough data to read the clientDir field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT21 = 0x00001124 #There is not enough data to read the clientTimeZone field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT22 = 0x00001125 #There is not enough data to read the clientSessionId field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	SECURITYDATATOOSHORT23 = 0x00001126 #There is not enough data to read the Client Info PDU Data (section 2.2.1.11.1).
	BADMONITORDATA = 0x00001129 #The number of TS_MONITOR_DEF (section 2.2.1.3.6.1) structures present in the monitorDefArray field of the Client Monitor Data (section 2.2.1.3.6) is less than the value specified in monitorCount field.
	VCDECOMPRESSEDREASSEMBLEFAILED = 0x0000112A #The server-side decompression buffer is invalid, or the size of the decompressed VC data exceeds the chunking size specified in the Virtual Channel Capability Set (section 2.2.7.1.10).
	VCDATATOOLONG = 0x0000112B #The size of a received Virtual Channel PDU (section 2.2.6.1) exceeds the chunking size specified in the Virtual Channel Capability Set (section 2.2.7.1.10).
	BAD_FRAME_ACK_DATA = 0x0000112C #There is not enough data to read a TS_FRAME_ACKNOWLEDGE_PDU ([MS-RDPRFX] section 2.2.3.1).
	GRAPHICSMODENOTSUPPORTED = 0x0000112D #The graphics mode requested by the client is not supported by the server.
	GRAPHICSSUBSYSTEMRESETFAILED = 0x0000112E #The server-side graphics subsystem failed to reset.
	GRAPHICSSUBSYSTEMFAILED = 0x0000112F #The server-side graphics subsystem is in an error state and unable to continue graphics encoding.
	TIMEZONEKEYNAMELENGTHTOOSHORT = 0x00001130 #There is not enough data to read the cbDynamicDSTTimeZoneKeyName field in the Extended Info Packet (section 2.2.1.11.1.1.1).
	TIMEZONEKEYNAMELENGTHTOOLONG = 0x00001131 #The length reported in the cbDynamicDSTTimeZoneKeyName field of the Extended Info Packet (section 2.2.1.11.1.1.1) is too long.
	DYNAMICDSTDISABLEDFIELDMISSING = 0x00001132 #The dynamicDaylightTimeDisabled field is not present in the Extended Info Packet (section 2.2.1.11.1.1.1).
	VCDECODINGERROR = 0x00001133 #An error occurred when processing dynamic virtual channel data ([MS-RDPEDYC] section 3.3.5).
	VIRTUALDESKTOPTOOLARGE = 0x00001134 #The width or height of the virtual desktop defined by the monitor layout in the Client Monitor Data (section 2.2.1.3.6) is larger than the maximum allowed value of 32,766.
	MONITORGEOMETRYVALIDATIONFAILED = 0x00001135 #The monitor geometry defined by the Client Monitor Data (section 2.2.1.3.6) is invalid.
	INVALIDMONITORCOUNT = 0x00001136 #The monitorCount field in the Client Monitor Data (section 2.2.1.3.6) is too large.
	UPDATESESSIONKEYFAILED = 0x00001191 #An attempt to update the session keys while using Standard RDP Security mechanisms (section 5.3.7) failed.
	DECRYPTFAILED = 0x00001192 #One of two possible error conditions: Decryption using Standard RDP Security mechanisms (section 5.3.6) failed. #Session key creation using Standard RDP Security mechanisms (section 5.3.5) failed.
	ENCRYPTFAILED = 0x00001193 #Encryption using Standard RDP Security mechanisms (section 5.3.6) failed.
	ENCPKGMISMATCH = 0x00001194 #Failed to find a usable Encryption Method (section 5.3.2) in the encryptionMethods field of the Client Security Data (section 2.2.1.4.3).
	DECRYPTFAILED2 = 0x00001195 #Unencrypted data was encountered in a protocol stream which is meant to be encrypted with Standard RDP Security mechanisms (section 5.3.6).

class TS_SET_ERROR_INFO_PDU:
	def __init__(self):
		self.shareDataHeader:TS_SHAREDATAHEADER = None
		self.errorInfoRaw:int = None
		self.errorInfo:ERRINFO = None

	def to_bytes(self):
		t = self.shareID.to_bytes(4, byteorder='little', signed = False)
		t += len(self.sourceDescriptor).to_bytes(2, byteorder='little', signed = False)
		t += (len(capdata) + 2 + 2).to_bytes(2, byteorder='little', signed = False)
		t += self.sourceDescriptor
		t += len(self.capabilitySets).to_bytes(2, byteorder='little', signed = False)
		t += self.pad2Octets.to_bytes(2, byteorder='little', signed = False)
		t += capdata
		t += self.sessionId.to_bytes(4, byteorder='little', signed = False)
		#self.shareControlHeader.totalLength = 6 + len(t)
		#t = self.shareControlHeader.to_bytes() + t
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SET_ERROR_INFO_PDU.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SET_ERROR_INFO_PDU()
		msg.shareDataHeader = TS_SHAREDATAHEADER.from_buffer(buff)
		msg.errorInfoRaw = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		try:
			msg.errorInfo = ERRINFO(msg.errorInfoRaw)
		except:
			msg.errorInfo = ERRINFO.UNK
		return msg

	def __repr__(self):
		t = '==== TS_SET_ERROR_INFO_PDU ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


from aardwolf.extensions.RDPECLIP.protocol.formatlist import CLIPBRD_FORMAT
from aardwolf.commons.queuedata import RDPDATATYPE


class RDP_CLIPBOARD_READY:
	"""
	This object will be dispatched on the external queue 
	when the server is ready to recieve/send clipboard commands
	"""
	def __init__(self):
		self.type = RDPDATATYPE.CLIPBOARD_READY

class RDP_CLIPBOARD_CONSUMED:
	"""
	This object will be dispatched on the external queue 
	when the client-side clipboard data has been consumed by the server
	"""
	def __init__(self):
		self.type = RDPDATATYPE.CLIPBOARD_CONSUMED
	
class RDP_CLIPBOARD_NEW_DATA_AVAILABLE:
	"""
	This object will be dispatched on the external queue 
	when there is new clipboard data available on the server.
	NOTE: this will not hold the data!
	"""
	def __init__(self):
		self.type = RDPDATATYPE.CLIPBOARD_NEW_DATA_AVAILABLE
	
class RDP_CLIPBOARD_DATA_TXT:
	"""
	This object will be dispatched on the external queue 
	when the new clipboard data has arrived from the server to our client
	"""
	def __init__(self):
		self.type = RDPDATATYPE.CLIPBOARD_DATA_TXT
		self.data = None
		self.datatype: CLIPBRD_FORMAT = None

	def get_data(self, fmt:CLIPBRD_FORMAT = None):
		if fmt is None:
			fmt = self.datatype
		if fmt == CLIPBRD_FORMAT.CF_UNICODETEXT:
			return self.data.encode('utf-16-le') + b'\x00'*3
		elif fmt == CLIPBRD_FORMAT.CF_TEXT:
			return self.data.encode('ascii', 'ignore') + b'\x00'
		elif fmt == CLIPBRD_FORMAT.CF_OEMTEXT:
			return self.data.encode('cp1252', 'ignore') + b'\x00'*3
		else:
			return self.data
	
	def __eq__(self, obj):
		if isinstance(obj, RDP_CLIPBOARD_DATA_TXT) is False:
			return False
		if self.data != obj.data or self.datatype != obj.datatype:
			return False
		return True



import io
import enum

class ORDERFLGS_SUPP(enum.IntFlag):
	TS_NEG_DSTBLT_INDEX = 0x01 #DstBlt Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.1).
	TS_NEG_PATBLT_INDEX = 0x02 #PatBlt Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.3) and OpaqueRect Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.5).
	TS_NEG_SCRBLT_INDEX = 0x04 #ScrBlt Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.7).<26>
	TS_NEG_MEMBLT_INDEX = 0x08 #MemBlt Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.9).<27>
	TS_NEG_MEM3BLT_INDEX = 0x10 #Mem3Blt Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.10).
	UnusedIndex1 = 0x20 #The contents of the byte at this index MUST be ignored.
	UnusedIndex2 = 0x40 #The contents of the byte at this index MUST be ignored.
	TS_NEG_DRAWNINEGRID_INDEX = 0x80 #DrawNineGrid Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.21).
	TS_NEG_LINETO_INDEX = 0x100 #LineTo Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.11).
	TS_NEG_MULTI_DRAWNINEGRID_INDEX = 0x200 #MultiDrawNineGrid Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.22).
	UnusedIndex3 = 0x400 #The contents of the byte at this index MUST be ignored.
	TS_NEG_SAVEBITMAP_INDEX = 0x800 #SaveBitmap Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.12).
	UnusedIndex4 = 0x1000 #The contents of the byte at this index MUST be ignored.
	UnusedIndex5 = 0x2000 #The contents of the byte at this index MUST be ignored.
	UnusedIndex6 = 0x4000 #The contents of the byte at this index MUST be ignored.
	TS_NEG_MULTIDSTBLT_INDEX = 0x8000 #MultiDstBlt Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.2).
	TS_NEG_MULTIPATBLT_INDEX = 0x10000 #MultiPatBlt Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.4).
	TS_NEG_MULTISCRBLT_INDEX = 0x20000 #MultiScrBlt Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.8).
	TS_NEG_MULTIOPAQUERECT_INDEX = 0x40000 #MultiOpaqueRect Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.6).
	TS_NEG_FAST_INDEX_INDEX = 0x80000 #FastIndex Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.14).
	TS_NEG_POLYGON_SC_INDEX = 0x100000 #PolygonSC Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.16) and PolygonCB Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.17).
	TS_NEG_POLYGON_CB_INDEX = 0x200000 #PolygonCB Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.17) and PolygonSC Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.16).
	TS_NEG_POLYLINE_INDEX = 0x400000 #Polyline Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.18).
	UnusedIndex7 = 0x800000 #The contents of the byte at this index MUST be ignored.
	TS_NEG_FAST_GLYPH_INDEX = 0x1000000 #FastGlyph Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.15).
	TS_NEG_ELLIPSE_SC_INDEX = 0x2000000 #EllipseSC Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.19) and EllipseCB Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.20).
	TS_NEG_ELLIPSE_CB_INDEX = 0x4000000 #EllipseCB Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.20) and EllipseSC Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.19).
	TS_NEG_INDEX_INDEX = 0x8000000 #GlyphIndex Primary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.1.2.13).
	UnusedIndex8 = 0x10000000 #The contents of the byte at this index MUST be ignored.
	UnusedIndex9 = 0x20000000 #The contents of the byte at this index MUST be ignored.
	UnusedIndex10 = 0x40000000 #The contents of the byte at this index MUST be ignored.
	UnusedIndex11 = 0x80000000 #The contents of the byte at this index MUST be ignored.
	
class ORDERFLAGS_EX(enum.IntFlag):
	CACHE_BITMAP_REV3_SUPPORT = 0x0002 #The Cache Bitmap (Revision 3) Secondary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.2.8) is supported.
	ALTSEC_FRAME_MARKER_SUPPORT = 0x0004 #The Frame Marker Alternate Secondary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.3.7) is supported.

class ORDERFLAG(enum.IntFlag):
	NEGOTIATEORDERSUPPORT = 0x0002 #Indicates support for specifying supported drawing orders in the orderSupport field. This flag MUST be set.
	ZEROBOUNDSDELTASSUPPORT = 0x0008 #Indicates support for the TS_ZERO_BOUNDS_DELTAS (0x20) flag ([MS-RDPEGDI] section 2.2.2.2.1.1.2). The client MUST set this flag.
	COLORINDEXSUPPORT = 0x0020 #Indicates support for sending color indices (not RGB values) in orders.
	SOLIDPATTERNBRUSHONLY = 0x0040 #Indicates that this party can receive only solid and pattern brushes.
	ORDERFLAGS_EXTRA_FLAGS = 0x0080 #Indicates that the orderSupportExFlags field contains valid data.

class TS_ORDER_CAPABILITYSET:
	def __init__(self):
		self.terminalDescriptor:bytes = b'\x00' * 16
		self.pad4octetsA:bytes = b'\x00' * 4
		self.desktopSaveXGranularity:int = 1
		self.desktopSaveYGranularity:int = 20
		self.pad2octetsA:bytes = b'\x00' * 2
		self.maximumOrderLevel:int = 1
		self.numberFonts:int = 0
		self.orderFlags:ORDERFLAG = None
		self.orderSupport:ORDERFLGS_SUPP = 0
		self.textFlags:bytes = b'\x00' * 2
		self.orderSupportExFlags:ORDERFLAGS_EX = 0
		self.pad4octetsB:bytes = b'\x00' * 4
		self.desktopSaveSize:int = 230400 #[MS-RDPEGDI] section 2.2.2.2.1.1.2.12
		self.pad2octetsC:bytes = b'\x00' * 2
		self.pad2octetsD:bytes = b'\x00' * 2
		self.textANSICodePage:bytes = b'\x00' * 2 #ignored
		self.pad2octetsE:bytes = b'\x00' * 2


	def to_bytes(self):
		t  = self.terminalDescriptor
		t += self.pad4octetsA
		t += self.desktopSaveXGranularity.to_bytes(2, byteorder='little', signed = False)
		t += self.desktopSaveYGranularity.to_bytes(2, byteorder='little', signed = False)
		t += self.pad2octetsA
		t += self.maximumOrderLevel.to_bytes(2, byteorder='little', signed = False)
		t += self.numberFonts.to_bytes(2, byteorder='little', signed = False)
		t += self.orderFlags.to_bytes(2, byteorder='little', signed = False)
		for bit in bin(self.orderSupport)[2:].ljust(32,'0')[::-1]:
			if bit == '1':
				t += b'\x01'
			else:
				t += b'\x00'
		t += self.textFlags
		t += self.orderSupportExFlags.to_bytes(2, byteorder='little', signed = False)
		t += self.pad4octetsB
		t += self.desktopSaveSize.to_bytes(4, byteorder='little', signed = False)
		t += self.pad2octetsC
		t += self.pad2octetsD
		t += self.textANSICodePage
		t += self.pad2octetsE
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_ORDER_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_ORDER_CAPABILITYSET()
		msg.terminalDescriptor = buff.read(16)
		msg.pad4octetsA = buff.read(4)
		msg.desktopSaveXGranularity = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.desktopSaveYGranularity = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.pad2octetsA = buff.read(2)
		msg.maximumOrderLevel = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.numberFonts = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.orderFlags = ORDERFLAG(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		#waaaarghhh
		temp = ''
		for _ in range(32):
			temp += str(int(bool(buff.read(1)[0])))
		msg.orderSupport = ORDERFLGS_SUPP(int(temp[::-1],2))
		msg.textFlags = buff.read(2)
		msg.orderSupportExFlags = ORDERFLAGS_EX(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.pad4octetsB = buff.read(4)
		msg.desktopSaveSize = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.pad2octetsC = buff.read(2)
		msg.pad2octetsD = buff.read(2)
		msg.textANSICodePage = buff.read(2)
		msg.pad2octetsE = buff.read(2)
		return msg

	def __repr__(self):
		t = '==== TS_ORDER_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
import enum

from aardwolf import logger
from aardwolf.protocol.pdu.input.keyboard import TS_KEYBOARD_EVENT
from aardwolf.protocol.pdu.input.unicode import TS_UNICODE_KEYBOARD_EVENT
from aardwolf.protocol.pdu.input.mouse import TS_POINTER_EVENT
from aardwolf.protocol.pdu.input.mousex import TS_POINTERX_EVENT
from aardwolf.protocol.pdu.input.sync import TS_SYNC_EVENT
from aardwolf.protocol.pdu.input.unused import TS_UNUSED_EVENT


class INPUT_EVENT(enum.Enum):
	SYNC = 0x0000 #Indicates a Synchronize Event (section 2.2.8.1.1.3.1.1.5).
	UNUSED = 0x0002 #Indicates an Unused Event (section 2.2.8.1.1.3.1.1.6).
	SCANCODE = 0x0004 #Indicates a Keyboard Event (section 2.2.8.1.1.3.1.1.1).
	UNICODE = 0x0005 #Indicates a Unicode Keyboard Event (section 2.2.8.1.1.3.1.1.2).
	MOUSE = 0x8001 #Indicates a Mouse Event (section 2.2.8.1.1.3.1.1.3).
	MOUSEX = 0x8002 #Indicates an Extended Mouse Event (section 2.2.8.1.1.3.1.1.4).

class TS_INPUT_EVENT:
	def __init__(self):
		self.eventTime: bytes = b'\x00'*4 #would have created a datetime obj but it is in fact ignroed by the server
		self.messageType:INPUT_EVENT = None
		self.slowPathInputData:bytes = None #always 6 bytes long

		# high level
		self.input = None
	
	@staticmethod
	def from_input(cap):
		t = TS_INPUT_EVENT()
		t.messageType = obj2otype[type(cap)]
		t.input = cap
		return t

	def to_bytes(self):
		t = self.eventTime
		t += self.messageType.value.to_bytes(2, byteorder='little', signed = False)
		t += self.input.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_INPUT_EVENT.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_INPUT_EVENT()
		msg.eventTime = buff.read(4)
		msg.messageType = INPUT_EVENT(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.slowPathInputData = buff.read(6)
		
		ot = otype2obj[msg.messageType]
		if ot is not None:
			msg.input = ot.from_bytes(msg.slowPathInputData)
		else:
			logger.debug('Not implemented parser! %s' % msg.capabilitySetType)
		return msg

	def __repr__(self):
		t = '==== TS_INPUT_EVENT ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

otype2obj = {
	INPUT_EVENT.SYNC : TS_SYNC_EVENT,
	INPUT_EVENT.UNUSED : TS_UNUSED_EVENT,
	INPUT_EVENT.SCANCODE : TS_KEYBOARD_EVENT,
	INPUT_EVENT.UNICODE : TS_UNICODE_KEYBOARD_EVENT,
	INPUT_EVENT.MOUSE : TS_POINTER_EVENT,
	INPUT_EVENT.MOUSEX : TS_POINTERX_EVENT,

}

obj2otype = {v: k for k, v in otype2obj.items()}



import enum
import io
from typing import List
from aardwolf.protocol.fastpath.palette import TS_UPDATE_PALETTE_DATA
from aardwolf.protocol.fastpath.bitmap import TS_UPDATE_BITMAP_DATA
from aardwolf.protocol.fastpath.pointer import TS_FP_LARGEPOINTERATTRIBUTE, TS_FP_CACHEDPOINTERATTRIBUTE, TS_FP_POINTERATTRIBUTE, TS_POINTERPOSATTRIBUTE, TS_FP_COLORPOINTERATTRIBUTE
from aardwolf.protocol.fastpath.surface import TS_SURFCMD
#from aardwolf.protocol.fastpath.order import TS_FP_UPDATE_ORDERS

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/68b5ee54-d0d5-4d65-8d81-e1c4025f7597
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/a1c4caa8-00ed-45bb-a06e-5177473766d3

class FASTPATH_ACTION(enum.Enum):
	FASTPATH = 0x0 #Indicates that the PDU is a fast-path output PDU.
	X224 = 0x3 #Indicates the presence of a TPKT Header ([T123] section 8) initial version byte which indicates that the PDU is a slow-path output PDU (in this case the full value of the initial byte MUST be 0x03).

class FASTPATH_SEC(enum.IntFlag):
	NONE = 0x0
	SECURE_CHECKSUM = 0x1 #Indicates that the MAC signature for the PDU was generated using the "salted MAC generation" technique (section 5.3.6.1.1). If this bit is not set, then the standard technique was used (sections 2.2.8.1.1.2.2 and 2.2.8.1.1.2.3).
	ENCRYPTED = 0x2 #Indicates that the PDU contains an 8-byte MAC signature after the optional length2 field (that is, the dataSignature field is present), and the contents of the PDU are encrypted using the negotiated encryption package (sections 5.3.2 and 5.3.6).

class TS_FP_FIPS_INFO:
	def __init__(self):
		self.length:int = 16
		self.version:int = 1
		self.padlen:int = None 

	def to_bytes(self):
		t = self.length.to_bytes(2, byteorder='little', signed=False)
		t += self.version.to_bytes(1, byteorder='little', signed=False)
		t += self.padlen.to_bytes(1, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FP_FIPS_INFO.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FP_FIPS_INFO()
		msg.length = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.version = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		msg.padlen = int.from_bytes(buff.read(1), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_FP_FIPS_INFO ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class FASTPATH_UPDATETYPE(enum.IntFlag):
	ORDERS = 0x0 #Indicates a Fast-Path Orders Update ([MS-RDPEGDI] section 2.2.2.2).
	BITMAP = 0x1 #Indicates a Fast-Path Bitmap Update (section 2.2.9.1.2.1.2).
	PALETTE = 0x2 #Indicates a Fast-Path Palette Update (section 2.2.9.1.2.1.1).
	SYNCHRONIZE = 0x3 #Indicates a Fast-Path Synchronize Update (section 2.2.9.1.2.1.3).
	SURFCMDS = 0x4 #Indicates a Fast-Path Surface Commands Update (section 2.2.9.1.2.1.10).
	PTR_NULL = 0x5 #Indicates a Fast-Path System Pointer Hidden Update (section 2.2.9.1.2.1.5).
	PTR_DEFAULT = 0x6 #Indicates a Fast-Path System Pointer Default Update (section 2.2.9.1.2.1.6).
	PTR_POSITION = 0x8 #Indicates a Fast-Path Pointer Position Update (section 2.2.9.1.2.1.4).
	COLOR = 0x9 #Indicates a Fast-Path Color Pointer Update (section 2.2.9.1.2.1.7).
	CACHED = 0xA #Indicates a Fast-Path Cached Pointer Update (section 2.2.9.1.2.1.9).
	POINTER = 0xB #Indicates a Fast-Path New Pointer Update (section 2.2.9.1.2.1.8).
	LARGE_POINTER = 0xC #Indicates a Fast-Path Large Pointer Update (section 2.2.9.1.2.1.11).

class FASTPATH_FRAGMENT(enum.Enum):
	SINGLE = 0x0 #The fast-path data in the updateData field is not part of a sequence of fragments.
	LAST = 0x1 #The fast-path data in the updateData field contains the last fragment in a sequence of fragments.
	FIRST = 0x2 #The fast-path data in the updateData field contains the first fragment in a sequence of fragments.
	NEXT = 0x3 #The fast-path data in the updateData field contains the second or subsequent fragment in a sequence of fragments.

class FASTPATH_OUTPUT_COMPRESSION(enum.IntFlag):
	USED = 0x2 #Indicates that the compressionFlags field is present. 

class TS_FP_UPDATE:
	def __init__(self):
		# HDR start
		self.updateCode: FASTPATH_UPDATETYPE = None
		self.fragmentation: FASTPATH_FRAGMENT = None
		self.compression: int = None
		# HDR end
		self.compressionFlags:int = None
		self.size:int = None
		self.updateData:bytes = None

		# high level
		self.update = None

	def to_bytes(self):
		data = self.update.to_bytes()
		t = (self.updateCode | self.fragmentation.value << 4 | self.compression << 6).to_bytes(1, byteorder='little', signed = False)
		t += self.compressionFlags.to_bytes(1, byteorder='little', signed = False)
		t += len(data).to_bytes(2, byteorder='little', signed = False)
		t += data
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FP_UPDATE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FP_UPDATE()
		updateHeader = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.updateCode = FASTPATH_UPDATETYPE(updateHeader & 0xF)
		msg.fragmentation = FASTPATH_FRAGMENT((updateHeader >> 4) & 3)
		msg.compression = FASTPATH_OUTPUT_COMPRESSION(updateHeader >> 6)
		if FASTPATH_OUTPUT_COMPRESSION.USED in msg.compression:
			print('Bulk compression not implemented!')
			msg.compressionFlags = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		msg.size = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.updateData = buff.read(msg.size)
		ot = otype2obj[msg.updateCode]
		if ot is not None:
			msg.update = ot.from_bytes(msg.updateData)
		
		return msg

	def __repr__(self):
		t = '==== TS_FP_UPDATE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

otype2obj = {
	FASTPATH_UPDATETYPE.ORDERS : None, # TODO
	FASTPATH_UPDATETYPE.BITMAP : TS_UPDATE_BITMAP_DATA,
	FASTPATH_UPDATETYPE.PALETTE : TS_UPDATE_PALETTE_DATA,
	FASTPATH_UPDATETYPE.SYNCHRONIZE : None, # this is normal
	FASTPATH_UPDATETYPE.SURFCMDS : TS_SURFCMD,
	FASTPATH_UPDATETYPE.PTR_NULL : None, # this is normal
	FASTPATH_UPDATETYPE.PTR_DEFAULT : None, # this is normal
	FASTPATH_UPDATETYPE.PTR_POSITION : TS_POINTERPOSATTRIBUTE,
	FASTPATH_UPDATETYPE.COLOR : TS_FP_COLORPOINTERATTRIBUTE,
	FASTPATH_UPDATETYPE.CACHED : TS_FP_CACHEDPOINTERATTRIBUTE,
	FASTPATH_UPDATETYPE.POINTER : TS_FP_POINTERATTRIBUTE,
	FASTPATH_UPDATETYPE.LARGE_POINTER : TS_FP_LARGEPOINTERATTRIBUTE,
}

obj2otype = {v: k for k, v in otype2obj.items()}

class TS_FP_UPDATE_PDU:
	def __init__(self):
		self.action:FASTPATH_ACTION = 0
		self.reserved:int = 0
		self.flags:FASTPATH_SEC = None
		self.length1:int = None
		self.length2:int = None
		self.fipsInformation:TS_FP_FIPS_INFO = None #OPTIONAL!
		self.dataSignature:bytes = None
		self.fpOutputUpdates:List[TS_FP_UPDATE] = []

	def to_bytes(self):
		raise NotImplementedError()

	@staticmethod
	def from_bytes(bbuff: bytes, with_fips = False):
		return TS_FP_UPDATE_PDU.from_buffer(io.BytesIO(bbuff), with_fips)

	@staticmethod
	def from_buffer(buff: io.BytesIO, with_fips = False):
		msg = TS_FP_UPDATE_PDU()
		t = buff.read(1)[0]
		msg.action = FASTPATH_ACTION(t & 3)
		msg.reserved = (t >> 2) & 15
		msg.flags = FASTPATH_SEC(t >> 6)


		msg.length1 = buff.read(1)[0]
		length = msg.length1
		if msg.length1 >> 7 == 1:
			msg.length2 = buff.read(1)[0]
			buff.seek(-2,1)
			length = int.from_bytes(buff.read(2), byteorder='big', signed=False)
			length = length & 0x7fff

		if with_fips is True:
			msg.fipsInformation = TS_FP_FIPS_INFO.from_buffer(buff)
		if FASTPATH_SEC.ENCRYPTED in msg.flags or FASTPATH_SEC.SECURE_CHECKSUM in msg.flags:
			msg.dataSignature = buff.read(8)
		
		if FASTPATH_SEC.ENCRYPTED in msg.flags:
			msg.fpOutputUpdates = buff.read()
		else:
			msg.fpOutputUpdates = TS_FP_UPDATE.from_buffer(buff)

		return msg

	def __repr__(self):
		t = '==== TS_SHAREDATAHEADER ====\r\n'
		for k in self.__dict__:
			if k == 'fpOutputUpdates':
				value = self.__dict__[k][:100]
			elif isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k].name
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



from aardwolf.extensions.RDPEDYC.vchannels import VirtualChannelBase
import asyncio


class VchannelTEST(VirtualChannelBase):
	def __init__(self):
		VirtualChannelBase.__init__(self, 'DATATEST1')
	
	async def channel_init(self):
		print('DATATEST1 channel started!')
		return True, None

	async def channel_data_in(self, data):
		print('DATATEST1 channel data in: %s' % data)
		#await self.channel_data_out(data)

		await asyncio.sleep(5)
		await self.channel_data_out(b'AB'*5000)


import io
import enum
from typing import List
from aardwolf.protocol.T128.share import TS_SHAREDATAHEADER

class TS_BITMAPCACHE_PERSISTENT_LIST_ENTRY:
	def __init__(self):
		self.Key1:int = 0
		self.Key2:int = 0

	def to_bytes(self):
		t = self.Key1.to_bytes(4, byteorder='little', signed = False)
		t += self.Key2.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAPCACHE_PERSISTENT_LIST_ENTRY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAPCACHE_PERSISTENT_LIST_ENTRY()
		msg.Key1 = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		msg.Key2 = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		return msg

	def __repr__(self):
		t = '==== TS_BITMAPCACHE_PERSISTENT_LIST_ENTRY ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_BITMAPCACHE_PERSISTENT_LIST_PDU:
	def __init__(self):
		self.shareDataHeader:TS_SHAREDATAHEADER = None
		self.numEntriesCache0:int = 0
		self.numEntriesCache1:int = 0
		self.numEntriesCache2:int = 0
		self.numEntriesCache3:int = 0
		self.numEntriesCache4:int = 0
		self.totalEntriesCache0:int = 0
		self.totalEntriesCache1:int = 0
		self.totalEntriesCache2:int = 0
		self.totalEntriesCache3:int = 0
		self.totalEntriesCache4:int = 0
		self.bBitMask:int = 0
		self.Pad2:int = 0
		self.Pad3:int = 0
		self.entries:List[TS_BITMAPCACHE_PERSISTENT_LIST_ENTRY] = 0


	def to_bytes(self):
		#t  = self.shareDataHeader.to_bytes()
		t = self.numberEntries.to_bytes(2, byteorder='little', signed = False)
		t += self.totalNumEntries.to_bytes(2, byteorder='little', signed = False)
		t += self.mapFlags.to_bytes(2, byteorder='little', signed = False)
		t += self.entrySize.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_BITMAPCACHE_PERSISTENT_LIST_PDU.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_BITMAPCACHE_PERSISTENT_LIST_PDU()
		msg.shareDataHeader = TS_SHAREDATAHEADER.from_buffer(buff)
		msg.numEntriesCache0 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.numEntriesCache1 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.numEntriesCache2 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.numEntriesCache3 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.numEntriesCache4 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.totalEntriesCache0 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.totalEntriesCache1 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.totalEntriesCache2 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.totalEntriesCache3 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.totalEntriesCache4 = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.bBitMask:int = 0
		msg.Pad2:int = 0
		msg.Pad3:int = 0
		msg.entries:List[TS_BITMAPCACHE_PERSISTENT_LIST_ENTRY] = 0



		return msg

	def __repr__(self):
		t = '==== TS_BITMAPCACHE_PERSISTENT_LIST_PDU ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t




# https://stackoverflow.com/questions/1094841/get-human-readable-version-of-file-size
def sizeof_fmt(num, suffix='B'):
	if num is None:
		return ''
	for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
		if abs(num) < 1024.0:
			return "%3.1f%s%s" % (num, unit, suffix)
		num /= 1024.0
	return "%.1f%s%s" % (num, 'Yi', suffix)


import io
import enum

class SURFCMDS(enum.IntFlag):
	SETSURFACEBITS = 0x00000002 #The Set Surface Bits Command (section 2.2.9.2.1) is supported.
	FRAMEMARKER = 0x00000010 #The Frame Marker Command (section 2.2.9.2.3) is supported.
	STREAMSURFACEBITS = 0x00000040 #The Stream Surface Bits Command (section 2.2.9.2.2) is supported.

class TS_SURFCMDS_CAPABILITYSET:
	def __init__(self):
		self.cmdFlags:SURFCMDS = None
		self.reserved:bytes = b'\x00'*4

	def to_bytes(self):
		t = self.cmdFlags.to_bytes(4, byteorder='little', signed = False)
		t += self.reserved
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SURFCMDS_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SURFCMDS_CAPABILITYSET()
		msg.cmdFlags = SURFCMDS(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		msg.reserved = buff.read(4)
		return msg

	def __repr__(self):
		t = '==== TS_SURFCMDS_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import io
from aardwolf.extensions.RDPEDYC.protocol import DYNVC_CMD, dynvc_header_to_bytes, dynvc_header_from_buff

class DYNVC_CAPS_REQ:
    def __init__(self):
        self.cbid:int = 0
        self.sp:int   = 0
        self.cmd:DYNVC_CMD = DYNVC_CMD.CAPS_RSP
        self.pad:int     = 0
        self.version:int = 1
        self.PriorityCharge0:int = None
        self.PriorityCharge1:int = None
        self.PriorityCharge2:int = None
        self.PriorityCharge3:int = None

    @staticmethod
    def from_bytes(data: bytes):
        return DYNVC_CAPS_REQ.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        msg = DYNVC_CAPS_REQ()
        msg.cbid, msg.sp, msg.cmd = dynvc_header_from_buff(buff)
        msg.pad = buff.read(1)[0]
        msg.version = int.from_bytes(buff.read(2), byteorder='little', signed=False)
        if msg.version > 1:
            msg.PriorityCharge0 = int.from_bytes(buff.read(2), byteorder='little', signed=False)
            msg.PriorityCharge1 = int.from_bytes(buff.read(2), byteorder='little', signed=False)
            msg.PriorityCharge2 = int.from_bytes(buff.read(2), byteorder='little', signed=False)
            msg.PriorityCharge3 = int.from_bytes(buff.read(2), byteorder='little', signed=False)
        return msg

    def to_bytes(self):
        t = b''
        t += dynvc_header_to_bytes(self.cbid, self.sp, self.cmd)
        t += bytes([self.pad])
        t += self.version.to_bytes(2, byteorder='little', signed=False)
        if self.version > 1:
            t += self.PriorityCharge0.to_bytes(2, byteorder='little', signed=False)
            t += self.PriorityCharge1.to_bytes(2, byteorder='little', signed=False)
            t += self.PriorityCharge2.to_bytes(2, byteorder='little', signed=False)
            t += self.PriorityCharge3.to_bytes(2, byteorder='little', signed=False)
        return t
    
    def __str__(self):
        t = ''
        for k in self.__dict__:
            t += '%s: %s\r\n' % (k, self.__dict__[k])
        return t


class DYNVC_CAPS_RSP:
    def __init__(self):
        self.cbid:int = 0
        self.sp:int   = 0
        self.cmd:DYNVC_CMD = DYNVC_CMD.CAPS_RSP
        self.pad:int     = 0
        self.version:int = 1

    @staticmethod
    def from_bytes(data: bytes):
        return DYNVC_CAPS_RSP.from_buffer(io.BytesIO(data))

    @staticmethod
    def from_buffer(buff: io.BytesIO):
        msg = DYNVC_CAPS_RSP()
        msg.cbid, msg.sp, msg.cmd = dynvc_header_from_buff(buff)
        msg.pad = buff.read(1)[0]
        msg.version = int.from_bytes(buff.read(2), byteorder='little', signed=False)
        return msg

    def to_bytes(self):
        t = b''
        t += dynvc_header_to_bytes(self.cbid, self.sp, self.cmd)
        t += bytes([self.pad])
        t += self.version.to_bytes(2, byteorder='little', signed=False)
        return t
    
    def __str__(self):
        t = ''
        for k in self.__dict__:
            t += '%s: %s\r\n' % (k, self.__dict__[k])
        return t


import io
import enum
from typing import List

class TS_POINT16:
	def __init__(self):
		self.xPos:int = None
		self.yPos:int = None

	def to_bytes(self):
		t = self.xPos.to_bytes(2, byteorder='little', signed=False)
		t += self.yPos.to_bytes(2, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_POINT16.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_POINT16()
		msg.xPos = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.yPos = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_POINT16 ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_FP_CACHEDPOINTERATTRIBUTE:
	def __init__(self):
		self.cachedPointerUpdateData:int = None 

	def to_bytes(self):
		t = self.cachedPointerUpdateData.to_bytes(2, byteorder='little', signed=False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FP_CACHEDPOINTERATTRIBUTE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FP_CACHEDPOINTERATTRIBUTE()
		msg.cachedPointerUpdateData = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		return msg

	def __repr__(self):
		t = '==== TS_FP_CACHEDPOINTERATTRIBUTE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_FP_POINTERATTRIBUTE:
	def __init__(self):
		self.xorBpp:int = None 
		self.colorPtrAttr:TS_FP_COLORPOINTERATTRIBUTE = None 

	def to_bytes(self):
		t = self.xorBpp.to_bytes(2, byteorder='little', signed=False)
		t += self.colorPtrAttr.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FP_POINTERATTRIBUTE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FP_POINTERATTRIBUTE()
		msg.xorBpp = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.colorPtrAttr = TS_FP_COLORPOINTERATTRIBUTE.from_buffer(buff)
		return msg

	def __repr__(self):
		t = '==== TS_FP_POINTERATTRIBUTE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_FP_LARGEPOINTERATTRIBUTE:
	def __init__(self):
		self.xorBpp:int = None
		self.cacheIndex:int = None
		self.hotSpot:TS_POINT16 = None
		self.width:int = None
		self.height:int = None
		self.lengthAndMask:int = None
		self.lengthXorMask:int = None
		self.xorMaskData:bytes = None
		self.andMaskData:bytes = None
		self.pad: bytes = None

	def to_bytes(self):
		t = self.xorBpp.to_bytes(2, byteorder='little', signed=False)
		t += self.cacheIndex.to_bytes(2, byteorder='little', signed=False)
		t += self.hotSpot.to_bytes()
		t += self.width.to_bytes(2, byteorder='little', signed=False)
		t += self.height.to_bytes(2, byteorder='little', signed=False)
		t += len(self.andMaskData).to_bytes(4, byteorder='little', signed=False)
		t += len(self.xorMaskData).to_bytes(4, byteorder='little', signed=False)
		t += self.xorMaskData
		t += self.andMaskData
		if len(t) % 2 != 0:
			t += b'\x00'
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FP_LARGEPOINTERATTRIBUTE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FP_LARGEPOINTERATTRIBUTE()
		msg.xorBpp = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.cacheIndex = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.hotSpot = TS_POINT16.from_buffer(buff)
		msg.width = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.height = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.lengthAndMask = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		msg.lengthXorMask = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		msg.xorMaskData = buff.read(msg.lengthAndMask)
		msg.andMaskData = buff.read(msg.lengthXorMask)
		msg.pad = buff.read(1)
		return msg

	def __repr__(self):
		t = '==== TS_FP_LARGEPOINTERATTRIBUTE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t

class TS_FP_COLORPOINTERATTRIBUTE:
	def __init__(self):
		self.cacheIndex:int = None
		self.hotSpot:TS_POINT16 = None
		self.width:int = None
		self.height:int = None
		self.lengthAndMask:int = None
		self.lengthXorMask:int = None
		self.xorMaskData:bytes = None
		self.andMaskData:bytes = None
		self.pad: bytes = None

	def to_bytes(self):
		t = self.cacheIndex.to_bytes(2, byteorder='little', signed=False)
		t += self.hotSpot.to_bytes()
		t += self.width.to_bytes(2, byteorder='little', signed=False)
		t += self.height.to_bytes(2, byteorder='little', signed=False)
		t += len(self.andMaskData).to_bytes(2, byteorder='little', signed=False)
		t += len(self.xorMaskData).to_bytes(2, byteorder='little', signed=False)
		t += self.xorMaskData
		t += self.andMaskData
		if len(t) % 2 != 0:
			t += b'\x00'
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_FP_COLORPOINTERATTRIBUTE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_FP_COLORPOINTERATTRIBUTE()
		msg.cacheIndex = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.hotSpot = TS_POINT16.from_buffer(buff)
		msg.width = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.height = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.lengthAndMask = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.lengthXorMask = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		msg.xorMaskData = buff.read(msg.lengthAndMask)
		msg.andMaskData = buff.read(msg.lengthXorMask)
		msg.pad = buff.read(1)
		return msg

	def __repr__(self):
		t = '==== TS_FP_COLORPOINTERATTRIBUTE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


class TS_POINTERPOSATTRIBUTE:
	def __init__(self):
		self.position:TS_POINT16 = None

	def to_bytes(self):
		t = self.position.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_POINTERPOSATTRIBUTE.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_POINTERPOSATTRIBUTE()
		msg.position = TS_POINT16.from_buffer(buff)
		return msg

	def __repr__(self):
		t = '==== TS_POINTERPOSATTRIBUTE ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t




import io
import enum

class OSMAJORTYPE(enum.Enum):
	UNSPECIFIED = 0x0000 #Unspecified platform
	WINDOWS = 0x0001 #Windows platform 
	OS2 = 0x0002 #OS/2 platform
	MACINTOSH = 0x0003 #Macintosh platform
	UNIX = 0x0004 #UNIX platform
	IOS= 0x0005 #iOS platform
	OSX = 0x0006 #OS X platform
	ANDROID = 0x0007 #Android platform
	CHROME_OS = 0x0008 #Chrome OS platform

class OSMINORTYPE(enum.Enum):
	UNSPECIFIED = 0x0000 #Unspecified version
	WINDOWS_31X = 0x0001 #Windows 3.1x
	WINDOWS_95 = 0x0002 #Windows 95
	WINDOWS_NT = 0x0003 #Windows NT
	OS2_V21 = 0x0004 #OS/2 2.1
	POWER_PC = 0x0005 #PowerPC
	MACINTOSH = 0x0006 #Macintosh
	NATIVE_XSERVER = 0x0007 #Native X Server
	PSEUDO_XSERVER = 0x0008 #Pseudo X Server
	WINDOWS_RT = 0x0009 #Windows RT

class EXTRAFLAG(enum.IntFlag):
	FASTPATH_OUTPUT_SUPPORTED = 0x0001 #Advertiser supports fast-path output.<24>
	NO_BITMAP_COMPRESSION_HDR = 0x0400 #Advertiser supports excluding the 8-byte Compressed Data Header (section 2.2.9.1.1.3.1.2.3) from the Bitmap Data (section 2.2.9.1.1.3.1.2.2) structure or the Cache Bitmap (Revision 2) Secondary Drawing Order ([MS-RDPEGDI] section 2.2.2.2.1.2.3).
	LONG_CREDENTIALS_SUPPORTED = 0x0004 #Advertiser supports long-length credentials for the user name, password, or domain name in the Save Session Info PDU (section 2.2.10.1).<25>
	AUTORECONNECT_SUPPORTED = 0x0008 #Advertiser supports auto-reconnection (section 5.5).
	ENC_SALTED_CHECKSUM = 0x0010 #Advertiser supports salted MAC generation (section 5.3.6.1.1).



class TS_GENERAL_CAPABILITYSET:
	def __init__(self):
		self.osMajorType:OSMAJORTYPE = None
		self.osMinorType:OSMINORTYPE = None
		self.protocolVersion:int = 0x0200
		self.pad2octetsA:bytes = b'\x00'*2
		self.compressionTypes:int = 0
		self.extraFlags:EXTRAFLAG = None
		self.updateCapabilityFlag:int = 0
		self.remoteUnshareFlag:int = 0
		self.compressionLevel:int = 0
		self.refreshRectSupport:bool = False
		self.suppressOutputSupport:bool = False

	def to_bytes(self):
		t = self.osMajorType.value.to_bytes(2, byteorder='little', signed = False)
		t += self.osMinorType.value.to_bytes(2, byteorder='little', signed = False)
		t += self.protocolVersion.to_bytes(2, byteorder='little', signed = False)
		t += self.pad2octetsA
		t += self.compressionTypes.to_bytes(2, byteorder='little', signed = False)
		t += self.extraFlags.to_bytes(2, byteorder='little', signed = False)
		t += self.updateCapabilityFlag.to_bytes(2, byteorder='little', signed = False)
		t += self.remoteUnshareFlag.to_bytes(2, byteorder='little', signed = False)
		t += self.compressionLevel.to_bytes(2, byteorder='little', signed = False)
		t += int(self.refreshRectSupport).to_bytes(1, byteorder='little', signed = False)
		t += int(self.suppressOutputSupport).to_bytes(1, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_GENERAL_CAPABILITYSET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_GENERAL_CAPABILITYSET()
		msg.osMajorType = OSMAJORTYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.osMinorType = OSMINORTYPE(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.protocolVersion = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.pad2octetsA = buff.read(2)
		msg.compressionTypes = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.extraFlags = EXTRAFLAG(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		msg.updateCapabilityFlag = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.remoteUnshareFlag = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.compressionLevel = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		msg.refreshRectSupport = bool(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		msg.suppressOutputSupport = bool(int.from_bytes(buff.read(1), byteorder='little', signed = False))
		return msg

	def __repr__(self):
		t = '==== TS_GENERAL_CAPABILITYSET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t



import enum
import io

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-rdpbcgr/ca73831d-3661-4700-9357-8f247640c02e1
class TS_SECURITY_PACKET:
	def __init__(self):
		self.length = None
		self.encryptedClientRandom:bytes = None

	def to_bytes(self):
		t = (len(self.encryptedClientRandom)).to_bytes(4, byteorder='little', signed=False)
		t += self.encryptedClientRandom
		return t
		#return self.encryptedClientRandom

	@staticmethod
	def from_bytes(bbuff: bytes):
		return TS_SECURITY_PACKET.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff: io.BytesIO):
		msg = TS_SECURITY_PACKET()
		msg.encryptedClientRandom = buff.read()
		return msg

	def __repr__(self):
		t = '==== TS_SECURITY_PACKET ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t




import asyncio
import traceback
from typing import List

from aardwolf import logger
from aardwolf.keyboard.layoutmanager import KeyboardLayoutManager
from aardwolf.keyboard import KeyboardLayout, VK_MODIFIERS

class DuckyExecutorBase:
	def __init__(self, keyboard_layout:KeyboardLayout, key_sender, send_as_char = False):
		self.keyboard_layout = keyboard_layout
		self.key_sender = key_sender
		self.send_as_char = send_as_char
		self.default_delay = 100
		self.default_chardelay = 50/1000
		self.aliases = {
			'do_default_delay' : 'do_defaultdelay',
			'do_windows' : 'do_gui',
			'do_app' : 'do_menu',
			'do_ctrl' : 'do_control',
			'do_uparrow' : 'do_up',
			'do_leftarrow' : 'do_left',
			'do_rightarrow' : 'do_right',
			'do_break' : 'do_pause',
			'do_delete' : 'do_del',
			'do_escape' : 'do_esc',
			'do_downarrow' : 'do_down',
			'do_delete' : 'do_del',
		}

		self.cmd_to_vk = {
			'backspace' : 'VK_BACK',
			'tab' : 'VK_TAB',
			'space' : 'VK_SPACE',
			'scrollock' : 'VK_SCROLL',
			'printscreen' : 'VK_SNAPSHOT',
			'pagedown' : 'VK_NEXT',
			'pageup' : 'VK_PRIOR',
			'numlock' : 'VK_NUMLOCK',
			'insert' : 'VK_INSERT',
			'home' : 'VK_HOME',
			'esc' : 'VK_ESCAPE',
			'escape' : 'VK_ESCAPE',
			'end' : 'VK_END',
			'del' : 'VK_DELETE',
			'delete' : 'VK_DELETE',
			'capslock' : 'VK_CAPITAL',
			'pause' : 'VK_PAUSE',
			'break' : 'VK_PAUSE',
			'right' : 'VK_RIGHT',
			'rightarrow' : 'VK_RIGHT',
			'left' : 'VK_LEFT',
			'leftarrow' : 'VK_LEFT',
			'down' : 'VK_DOWN',
			'downarrow' : 'VK_DOWN',
			'up' : 'VK_UP',
			'uparrow' : 'VK_UP',
			'gui': 'VK_LWIN',
			'windows': 'VK_LWIN',
			'enter' : 'VK_RETURN',
			'shift' : 'VK_LSHIFT',
			'alt' : 'VK_LMENU',
			'f1' : 'VK_F1',
			'f2' : 'VK_F2',
			'f3' : 'VK_F3',
			'f4' : 'VK_F4',
			'f5' : 'VK_F5',
			'f6' : 'VK_F6',
			'f7' : 'VK_F7',
			'f8' : 'VK_F8',
			'f9' : 'VK_F9',
			'f10' : 'VK_F10',
			'f11' : 'VK_F11',
			'f12' : 'VK_F12',

		}
	
	def get_function(self, cmdname):
		if cmdname.startswith('do_') is False:
			cmdname = 'do_' + cmdname
		
		if cmdname in self.aliases:
			cmdname = self.aliases[cmdname]
		
		return getattr(self, cmdname)

	async def do_enter(self):
		await self.keydispatch('VK_RETURN')
		#await asyncio.sleep(self.default_delay)

	async def do_function(self, data):
		await self.keydispatch('VK_F%s' % data[0])
	
	async def do_rem(self, data:List[str]):
		data = ' '.join(data)
		logger.debug(data)

	async def do_defaultdelay(self, delay:int):
		delay = ' '.join(delay)
		self.default_delay = int(delay) / 1000 #protocol allows delay to be set in 10 miliseconds interval
	
	async def do_defaultchardelay(self, delay:int):
		delay = ' '.join(delay)
		self.default_chardelay = int(delay) / 1000
	
	async def do_delay(self, delay:int):
		pass
		delay = ' '.join(delay)
		delay = int(delay) / 1000
		if delay > 1:
			return
		await asyncio.sleep(delay)

	async def do_string(self, data:str):
		data = ' '.join(data)
		if self.send_as_char is True:
			for c in data:
				await self.key_sender(c, True, True)
				await asyncio.sleep(self.default_chardelay)
		else:
			for c in data:
				await self.keydispatch(c)

	async def do_gui(self, data = []):
		if len(data) > 0:
			data.insert(0, 'VK_LWIN')
			await self.multi_key_press(data)
		else:
			await self.keydispatch('VK_LWIN')
		#await asyncio.sleep(self.default_delay)

	async def do_menu(self):
		await self.keydispatch('VK_RMENU')
		#await asyncio.sleep(self.default_delay)

	async def do_shift(self, data = []):
		print('SHIFT + %s' % repr(data))
		if len(data) > 0:
			data.insert(0, 'VK_LSHIFT')
			await self.multi_key_press(data)
		else:
			await self.keydispatch('VK_LSHIFT')

	async def do_control(self, data = []):
		print('CTRL + %s' % repr(data))
		if len(data) > 0:
			data.insert(0, 'VK_LCONTROL')
			await self.multi_key_press(data)
		else:
			await self.keydispatch('VK_LCONTROL')

	async def do_alt(self, data = []):
		print('ALT + %s' % repr(data))
		if len(data) > 0:
			data.insert(0, 'VK_LMENU')
			await self.multi_key_press(data)
		else:
			await self.keydispatch('VK_LMENU')

	async def do_up(self):
		await self.keydispatch('VK_UP')
		#await asyncio.sleep(self.default_delay)

	async def do_down(self):
		await self.keydispatch('VK_DOWN')
		#await asyncio.sleep(self.default_delay)

	async def do_left(self):
		await self.keydispatch('VK_LEFT')
		#await asyncio.sleep(self.default_delay)

	async def do_right(self):
		await self.keydispatch('VK_RIGHT')
		#await asyncio.sleep(self.default_delay)
	
	async def do_pause(self):
		await self.keydispatch('VK_PAUSE')
		#await asyncio.sleep(self.default_delay)
	
	async def do_capslock(self):
		await self.keydispatch('VK_CAPITAL')
		#await asyncio.sleep(self.default_delay)
	
	async def do_del(self):
		await self.keydispatch('VK_DELETE')
		#await asyncio.sleep(self.default_delay)

	async def do_end(self):
		await self.keydispatch('VK_END')
		#await asyncio.sleep(self.default_delay)
	
	async def do_esc(self):
		await self.keydispatch('VK_ESCAPE')
		#await asyncio.sleep(self.default_delay)

	async def do_home(self):
		await self.keydispatch('VK_HOME')
		#await asyncio.sleep(self.default_delay)
	
	async def do_insert(self):
		await self.keydispatch('VK_INSERT')
		#await asyncio.sleep(self.default_delay)
	
	async def do_numlock(self):
		await self.keydispatch('VK_NUMLOCK')
		#await asyncio.sleep(self.default_delay)
	
	async def do_pageup(self):
		await self.keydispatch('VK_PRIOR')
		#await asyncio.sleep(self.default_delay)
	
	async def do_pagedown(self):
		await self.keydispatch('VK_NEXT')
		#await asyncio.sleep(self.default_delay)
	
	async def do_printscreen(self):
		await self.keydispatch('VK_SNAPSHOT')
		#await asyncio.sleep(self.default_delay)
	
	async def do_scrollock(self):
		await self.keydispatch('VK_SCROLL')
		#await asyncio.sleep(self.default_delay)
	
	async def do_space(self):
		await self.keydispatch('VK_SPACE')
		#await asyncio.sleep(self.default_delay)
	
	async def do_tab(self):
		await self.keydispatch('VK_TAB')
		#await asyncio.sleep(self.default_delay)
	
	async def do_backspace(self):
		await self.keydispatch('VK_BACK')
		#await asyncio.sleep(self.default_delay)
	
	async def multi_key_press(self, keys):
		codes = []
		#print(keys)
		for x in keys:
			#print(x)
			if len(x) == 1:
				scancode, mo = self.keyboard_layout.char_to_scancode(x.lower())
				if mo != VK_MODIFIERS(0) and mo != VK_MODIFIERS.VK_CAPITAL|VK_MODIFIERS.VK_SHIFT:
					print(mo)
					raise Exception('This is not supported!!!!')
				codes.append(scancode)
			else:
				if x.lower() in self.cmd_to_vk:
					codes.append(self.keyboard_layout.vk_to_scancode(self.cmd_to_vk[x.lower()]))
				else:
					codes.append(self.keyboard_layout.vk_to_scancode(x))
		
		for code in codes:
			await asyncio.sleep(self.default_chardelay)
			await self.key_sender(code, True)
		
		for code in codes[::-1]:
			await asyncio.sleep(self.default_chardelay)
			await self.key_sender(code, False)
	
	async def keydispatch(self, key, modifiers = VK_MODIFIERS(0)):
		if key in '0123456789':
			key = 'VK_%s' % key
		if len(key) == 1:
			try:
				scancode, mo = self.keyboard_layout.char_to_scancode(key)
				#print('key     : %s' % key)
				#print('scancode: %s' % scancode)
				#print('mo      : %s' % mo)

			except KeyError:
				# this is bad...
				await asyncio.sleep(self.default_chardelay)
				await self.key_sender(key, True, True)
				return

			if mo == VK_MODIFIERS(0):
				await asyncio.sleep(self.default_chardelay)
				await self.key_sender(scancode, True)
				await asyncio.sleep(self.default_chardelay)
				await self.key_sender(scancode, False)
			else:
				#print('key: %s' % key)
				#print('sc : %s' % scancode)
				#print('mo : %s' % repr(mo))
				#input()
				modcodes = []
				afs = [flag for flag in VK_MODIFIERS if flag in mo]
				for af in afs:
					if af == VK_MODIFIERS.VK_CAPITAL:
						af = VK_MODIFIERS.VK_SHIFT
					modname = af.name
					if modname in ['VK_SHIFT', 'VK_CONTROL', 'VK_MENU']:
						modname = modname[:3] + 'L' + modname[3:]
					sc = self.keyboard_layout.vk_to_scancode(modname)
					modcodes.append(sc)
				for mod in modcodes:
					await asyncio.sleep(self.default_chardelay)
					await self.key_sender(mod, True)

				await asyncio.sleep(self.default_chardelay)
				await self.key_sender(scancode, True)
				await asyncio.sleep(self.default_chardelay)
				await self.key_sender(scancode, False)
				for mod in modcodes[::-1]:
					await asyncio.sleep(self.default_chardelay)
					await self.key_sender(mod, False)
				
		else:
			scancode = self.keyboard_layout.vk_to_scancode(key)
			await asyncio.sleep(self.default_chardelay)
			await self.key_sender(scancode, True)
			await asyncio.sleep(self.default_chardelay)
			await self.key_sender(scancode, False)
	

class DuckyReaderBase:
	def __init__(self, executor):
		self.executor = executor

	async def execute_line(self, line):
		try:
			line = line.strip()
			if line == '':
				return
			if line.find(' ') == -1:
				keyword = line.lower()
				data = []
			else:
				keyword, *data = line.split(' ')
				keyword = keyword.lower()
			
			if keyword.startswith('#') is True:
				return
			
			if keyword.startswith('rem'):
				if keyword[3:] != '':
					data.insert(0, keyword[3:])
				keyword = 'rem'
			
			if keyword.find('-') != -1:
				keyword, *temp = keyword.split('-')
				keyword = keyword.lower()
				data = temp + data
			
			if keyword.startswith('f') is True and (len(keyword) == 2 or len(keyword) == 3):
				keyid = int(keyword[1:])
				keyword = 'function'
				data.insert(0, str(keyid))
			fnct = self.executor.get_function(keyword)
			if len(data) > 0:
				await fnct(data)
			else:
				await fnct()
		
		except Exception as e:
			print('ERROR! Line: %s' % repr(line))
			raise e

	@staticmethod
	def from_file(filepath, executor):
		reader = DuckyReaderFile(filepath, executor)
		return reader

class DuckyReaderFile(DuckyReaderBase):
	def __init__(self, filepath, executor):
		DuckyReaderBase.__init__(self, executor)
		self.filepath = filepath
		self.file = open(filepath, 'r')

	async def parse(self):
		for line in self.file:
			line = line.strip()
			if len(line) == 0:
				continue
			await self.execute_line(line)


async def key_sender(scancode, is_pressed, as_char = False):
	#print(scancode)
	#print(is_pressed)
	return

async def amain():
	layout = KeyboardLayoutManager().get_layout_by_shortname('enus')
	from pathlib import Path
	try:
		for path in Path('/home/webdev/Desktop/usbrubberducky-payloads/payloads/').rglob('*'):
			print('Now processing %s' % path)
			if path.is_dir():
				continue
			if path.name.endswith('.md') is True or path.name.endswith('.bat') is True:
				continue
			executor = DuckyExecutorBase(layout, key_sender)
			reader = DuckyReaderFile.from_file(path, executor)
			await reader.parse()
	except Exception as e:
		traceback.print_exc()


def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()


import librlers
from PIL import Image

bpp_2_bytes = {
	16: 2,
	24: 3,
	32: 4,
}

def rectconvert(width, height, bitsPerPixel, isCompress, data):
	if bitsPerPixel not in bpp_2_bytes:
		raise ValueError("bitsPerPixel value of %s is not supported!" % bitsPerPixel)
	image = librlers.bitmap_decompress(data, width, height, 16, int(isCompress))
	return Image.frombytes('RGBA', [width, height], bytes(image))



import asyncio
import logging
from aardwolf.commons.factory import RDPConnectionFactory
from aardwolf.commons.queuedata.constants import VIDEO_FORMAT
from aardwolf.commons.iosettings import RDPIOSettings
from asysocks.unicomm.common.scanner.targetgen import UniTargetGen
from asysocks.unicomm.common.scanner.scanner import UniScanner
from aardwolf import logger

from asysocks.unicomm.common.target import UniTarget
from asyauth.common.credentials import UniCredential


from aardwolf.examples.scanners.rdplogin import RDPLoginScanner
from aardwolf.examples.scanners.rdpscaps import RDPCapabilitiesScanner
from aardwolf.examples.scanners.rdpscreen import RDPScreenshotScanner

rdpscan_options = {
	'login' : (RDPLoginScanner,'Checks if user can log in to hosts'),
	'caps'  : (RDPCapabilitiesScanner,'Lists RDP connection flags available'),
	'screen': (RDPScreenshotScanner,'Takes screenshot of the remote session'),
}

async def amain():
	import argparse

	protocols = """RDP : RDP protocol
	VNC: VNC protocol"""
	authprotos = """ntlm     : CREDSSP+NTLM authentication
	kerberos : CREDSSP+Kerberos authentication
	sspi-ntlm: CREDSSP+NTLM authentication using current user's creds (Windows only, restricted admin mode only)
	sspi-kerberos: CREDSSP+KERBEROS authentication using current user's creds (Windows only, restricted admin mode only)
	plain    : Old username and password authentication (only works when NLA is disabled on the server)
	none     : No authentication (same as plain, but no provided credentials needed)
	"""
	usage = UniCredential.get_help(protocols, authprotos, '')
	usage += UniTarget.get_help()
	usage += """
RDP Examples:
	Login with no credentials (only works when NLA is disabled on the server):
		rdp://10.10.10.2
	Login with username and password (only works when NLA is disabled on the server):
		rdp://TEST\Administrator:Passw0rd!1@10.10.10.2
	Login via CREDSSP+NTLM:
		rdp+ntlm-password://TEST\Administrator:Passw0rd!1@10.10.10.2
	Login via CREDSSP+Kerberos:
		rdp+kerberos-password://TEST\Administrator:Passw0rd!1@win2019ad.test.corp/?dc=10.10.10.2
	Login via CREDSSP+NTLM using current user's creds (Windows only, restricted admin mode only):
		rdp+sspi-ntlm://win2019ad.test.corp
	Login via CREDSSP+Kerberos using current user's creds (Windows only, restricted admin mode only):
		rdp+sspi-kerberos://win2019ad.test.corp/
	
VNC examples:
	Login with no credentials:
		vnc://10.10.10.2
	Login with password (the short way):
		vnc://Passw0rd!1@10.10.10.2
	Login with password:
		vnc+plain-password://Passw0rd!1@10.10.10.2
"""

	scannertpes_usage = '\r\nall: Runs all scanners\r\n'
	for k in rdpscan_options:
		scannertpes_usage += '    %s: %s\r\n' % (k, rdpscan_options[k][1])
	
	usage += """
Scanner types (-s param):
    %s
"""% scannertpes_usage

	parser = argparse.ArgumentParser(description='RDP scanner', usage=usage)
	parser.add_argument('-w', '--worker-count', type=int, default=100, help='Parallell count')
	parser.add_argument('-t', '--timeout', type=int, default=10, help='Timeout for each connection')
	parser.add_argument('--no-progress', action='store_false', help='Disable progress bar')
	parser.add_argument('-o', '--out-file', help='Output file path.')
	parser.add_argument('-s', '--scan', action='append', required=True, help='Scanner type')
	parser.add_argument('-e', '--errors', action='store_true', help='Includes errors in output. It will mess up the formatting!')
	parser.add_argument('url', help = 'Connection string in URL format')
	parser.add_argument('targets', nargs='*', help = 'Hostname or IP address or file with a list of targets')
	args = parser.parse_args()

	if len(args.targets) == 0:
		print('No targets defined!')
		return
	
	logger.setLevel(logging.INFO)

	iosettings = RDPIOSettings()
	iosettings.channels = []
	iosettings.video_out_format = VIDEO_FORMAT.RAW
	iosettings.clipboard_use_pyperclip = False
	
	connectionfactory = RDPConnectionFactory.from_url(args.url, iosettings)
	scantypes = []
	for x in args.scan:
		scantypes.append(x.lower())
	executors = []
	if 'all' in scantypes:
		for k in rdpscan_options:
			executors.append(rdpscan_options[k][0](connectionfactory))
	else:
		for scantype in scantypes:
			if scantype not in rdpscan_options:
				print('Unknown scan type: "%s"' % scantype)
				return
			executors.append(rdpscan_options[scantype][0](connectionfactory))
		
	tgen = UniTargetGen.from_list(args.targets)
	scanner = UniScanner('RDPScanner', executors, [tgen], worker_count=args.worker_count, host_timeout=args.timeout)
	await scanner.scan_and_process(progress=args.no_progress, out_file=args.out_file, include_errors=args.errors)

def main():
	asyncio.run(amain())

if __name__ == '__main__':
	main()


import enum
from aardwolf.commons.queuedata import RDPDATATYPE

class RDP_BEEP:
	"""
	This object will be dispatched on the external queue 
	when the server sends a Beep or Bell notification
	"""
	def __init__(self):
		self.type = RDPDATATYPE.BEEP
	def __repr__(self):
		t = '==== RDP_BEEP ====\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], enum.IntFlag):
				value = self.__dict__[k]
			elif isinstance(self.__dict__[k], enum.Enum):
				value = self.__dict__[k].name
			else:
				value = self.__dict__[k]
			t += '%s: %s\r\n' % (k, value)
		return t


import enum
from typing import List
from asysocks.unicomm.common.proxy import UniProxyProto, UniProxyTarget
from aardwolf.commons.iosettings import RDPIOSettings

from aardwolf.commons.target import RDPTarget, RDPConnectionDialect
from asyauth.common.credentials import UniCredential
from aardwolf.connection import RDPConnection
from aardwolf.vncconnection import VNCConnection
import ipaddress
import copy


class RDPConnectionFactory:
	def __init__(self, iosettings=None, target:RDPTarget = None, credential:UniCredential = None, proxies:List[UniProxyTarget] = None):
		self.credential = credential
		self.target = target
		self.proxies = proxies
		self.iosettings = iosettings

	def get_connection(self, iosettings):
		tio = copy.deepcopy(iosettings)
		credential = self.get_credential()
		target = self.get_target()
		if target.dialect is not None:
			if target.dialect == RDPConnectionDialect.RDP:
				return RDPConnection(target, credential, tio)
			elif target.dialect == RDPConnectionDialect.VNC:
				return VNCConnection(target, credential, tio)
			else:
				raise Exception('Unknown dialect %s' % target.dialect)
		raise Exception('Either target or dialect must be defined first!')

	def create_connection_newtarget(self, ip_or_hostname, iosettings):
		tio = copy.deepcopy(iosettings)
		credential = self.get_credential()

		target = self.get_target()
		try:
			ipaddress.ip_address(ip_or_hostname)
			target.ip = ip_or_hostname
			target.hostname = None
		except:
			target.hostname = ip_or_hostname
			target.ip = ip_or_hostname

		if target.dialect == RDPConnectionDialect.RDP:
			return RDPConnection(target, credential, tio)
		elif target.dialect == RDPConnectionDialect.VNC:
			return VNCConnection(target, credential, tio)
		else:
			raise Exception('Unknown dialect %s' % target.dialect)

	def get_proxy(self) -> List[UniProxyTarget]:
		return copy.deepcopy(self.target.proxies)

	def get_target(self) -> RDPTarget:
		return copy.deepcopy(self.target)	

	def get_credential(self) -> UniCredential:
		return copy.deepcopy(self.credential)
	
	def get_settings(self) -> RDPIOSettings:
		return copy.deepcopy(self.iosettings)

	@staticmethod
	def from_url(connection_url, iosettings):
		target = RDPTarget.from_url(connection_url)
		credential = UniCredential.from_url(connection_url)
		return RDPConnectionFactory(iosettings, target, credential)


	def __str__(self):
		t = '==== RDPConnectionFactory ====\r\n'
		for k in self.__dict__:
			val = self.__dict__[k]
			if isinstance(val, enum.IntFlag):
				val = val
			elif isinstance(val, enum.Enum):
				val = val.name
			
			t += '%s: %s\r\n' % (k, str(val))
			
		return t

	
			
if __name__ == '__main__':
	url_tests = [
		#'RDP://10.10.10.2',
		#'RDP://10.10.10.2:9000',
		#'RDP-tcp://10.10.10.2',
		#'RDP-tcp://10.10.10.2:9000',
		#'RDP-udp://10.10.10.2:138',
		#'RDP+ntlm-password://domain\\user@10.10.10.2',
		#'RDP-tcp+ntlm-password://domain\\user:password@10.10.10.2',
		#'RDP-tcp+ntlm-password://domain\\user:password@10.10.10.2:10000',
		#'RDP-tcp+ntlm-nt://domain\\user:alma@10.10.10.2',
		#'RDP+ntlm-nt://domain\\user:alma@10.10.10.2',
		#'RDP+ntlm-nt://domain\\user:alma@10.10.10.2',
		#'RDP-tcp+kerberos-password://domain\\alma:password@10.10.10.10.2',
		#'RDP-tcp+kerberos-aes://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'RDP-tcp+kerberos-aes://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'RDP-tcp+kerberos-nt://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'RDP-tcp+kerberos-rc4://domain\\alma:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa@10.10.10.10.2',
		#'RDP+sspi://10.10.10.10.2',
		#'RDP+sspi-ntlm://10.10.10.10.2',
		#'RDP+sspi-kerberos://10.10.10.10.2',
		#'RDP+multiplexor://10.10.10.10.2',
		#'RDP+multiplexor-ssl://10.10.10.10.2',
		#'RDP+multiplexor-ssl-ntlm://10.10.10.10.2',
		#'RDP+multiplexor-ssl-kerberos://10.10.10.10.2',
		#'RDP://10.10.10.2/?timeout=10',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyserver=127.0.0.1',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyserver=127.0.0.1&proxyuser=admin&proxypass=alma',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5&proxyserver=127.0.0.1&proxyuser=admin&proxypass=alma&dc=10.10.10.2&dns=8.8.8.8',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=socks5-ssl&proxyserver=127.0.0.1&proxyuser=admin&proxypass=alma&dc=10.10.10.2&dns=8.8.8.8',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=multiplexor&proxyserver=127.0.0.1',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=multiplexor&proxyserver=127.0.0.1&proxyagentid=alma',
		#'RDP://10.10.10.2/?timeout=10&dc=10.10.10.2&proxytype=multiplexor&proxyserver=127.0.0.1&proxyagentid=alma&proxytimeout=111',
		'RDP://10.10.10.2/C$/test/tst111.dmp?timeout=10&dc=10.10.10.2&proxytype=multiplexor&proxyhost=127.0.0.1&proxyport=1&proxyagentid=alma&proxytimeout=111',

	]
	for url in url_tests:
		print('===========================================================================')
		print(url)
		try:
			dec = RDPConnectionFactory.from_url(url)
			creds = dec.get_credential()
			target = dec.get_target()
			RDPfile = dec.get_file()
			print(RDPfile)
		except Exception as e:
			import traceback
			traceback.print_exc()
			print('ERROR! Reason: %s' % e)
			input()
		else:
			print(str(creds))
			print(str(target))
			input()
			
			
